from metamorf.engines.engine import Engine
from metamorf.tools.filecontroller import FileControllerFactory
from metamorf.tools.connection import ConnectionFactory
from metamorf.constants import *
import os

class EngineDeployExample(Engine):

    def _initialize_engine(self):
        self.engine_name = "Engine Deploy Example"
        self.engine_command = "deploy-example"


    def run(self):
        # Starts the execution loading the Configuration File. If there is an error it finishes the execution.
        super().start_execution()

        # METADATA DEPLOYMENT EXAMPLE
        connection_type = self.configuration_file['metadata']['connection_type']
        connection_type_data = self.configuration_file['data']['connection_type']
        # Get File that will be executed
        file_controller_init_sql = FileControllerFactory().get_file_reader(FILE_TYPE_SQL)
        file_controller_init_sql.set_file_location(os.path.join(PACKAGE_PATH, INITIALIZATION_FILE_PATH, connection_type.lower()), "init_example_metadata_" + connection_type_data.lower() + ".sql")
        file_init_sql = file_controller_init_sql.read_file()

        # Get Connection and Execute the Deployment
        self.log.log(self.engine_name, "Deploying Metamorf Metadata Example on the Database indicated on Configuration File", LOG_LEVEL_INFO)
        connection = ConnectionFactory().get_connection(connection_type)
        connection.setup_connection(self.configuration_file['metadata'], self.log)
        result = connection.execute(file_init_sql)
        if not result:
            self.log.log(self.engine_name, "The deployment can't be performed. Make sure Metamorf is installed correctly. You can do it using [deploy] command.", LOG_LEVEL_ERROR)
            super().finish_execution(False)
        connection.commit()
        self.log.log(self.engine_name, "Deployment Metamorf Metadata Example Finished", LOG_LEVEL_INFO)
        connection.close()

        # DATA DEPLOYMENT EXAMPLE
        connection_type = self.configuration_file['data']['connection_type']
        # Get File that will be executed
        file_controller_init_sql = FileControllerFactory().get_file_reader(FILE_TYPE_SQL)
        file_controller_init_sql.set_file_location(os.path.join(PACKAGE_PATH, INITIALIZATION_FILE_PATH, connection_type.lower()), "init_example_data_" + connection_type.lower() + ".sql")
        file_init_sql = file_controller_init_sql.read_file()

        # Get Connection and Execute the Deployment
        self.log.log(self.engine_name, "Deploying Metamorf Data Example on the Database indicated on Configuration File", LOG_LEVEL_INFO)
        connection = ConnectionFactory().get_connection(connection_type)
        connection.setup_connection(self.configuration_file['data'], self.log)
        result = connection.execute(file_init_sql)
        if not result: super().finish_execution(False)
        connection.commit()
        self.log.log(self.engine_name, "Deploying Metamorf Data Example Finished", LOG_LEVEL_INFO)
        connection.close()

        super().finish_execution()






from metamorf.engines.engine import Engine
from metamorf.constants import *
from metamorf.tools.manifest import generate_manifest
from metamorf.tools.utils import get_list_nodes_from_metadata

class EngineManifest(Engine):

    def _initialize_engine(self):
        self.engine_name = "Engine Manifest"
        self.engine_command = "manifest"

    def run(self):
        self.start_execution(need_configuration_file=True)
        self.metadata = self.load_metadata(load_om=True, load_entry=False, load_ref=True, load_im=True, owner=self.owner)
        self.log.log(self.engine_name, "Starting the create Manifest File", LOG_LEVEL_INFO)

        all_nodes = get_list_nodes_from_metadata(self.metadata, self.log, "all")
        generate_manifest(self.metadata, self.configuration_file, all_nodes)

        self.log.log(self.engine_name, "Manifest File Finished", LOG_LEVEL_INFO)
        self.finish_execution()







from metamorf.tools.filecontroller import FileControllerFactory
from metamorf.constants import *
import os

def print_help():
    file_controller_properties = FileControllerFactory().get_file_reader(FILE_TYPE_YML)
    file_controller_properties.set_file_location(os.path.join(PACKAGE_PATH, PROPERTIES_FILE_PATH), PROPERTIES_FILE_NAME)
    properties_file = file_controller_properties.read_file()

    print()
    print(properties_file['description'])
    print()
    print('usage: metamorf [command] [arguments]')
    print()
    print('commands:')

    for com in properties_file['commands']:
        all_args = []
        if 'args' in com:
            for opt in com['args']:
                for x in opt['options']:
                    all_args.append(x)
        all_args_string = ""
        for x in all_args: all_args_string += x + ", "
        all_args_string = all_args_string[:-2]

        if all_args_string == "": print("  " + com['name'] + " : " + com['description'])
        else: print("  " + com['name'] + " : " + com['description'] + " Optional arguments [" + all_args_string + "]")


import src.metamorf.utils as utils
import src.metamorf.tools as tools
import src.metamorf as metamorf

file = utils.FileControllerFactory().get_file_reader(metamorf.FILE_TYPE_YML)
file.set_file_location(r'C:\Users\guill\Documents\GitHub\metamorf\src\configuration', 'configuration.yml')
configurationFile = file.read_file()

file = utils.FileControllerFactory().get_file_reader(metamorf.FILE_TYPE_YML)
file.set_file_location(r'C:\Users\guill\Documents\GitHub\metamorf\src\configuration', 'properties.yml')
propertiesFile = file.read_file()

validator = tools.ConfigValidator(propertiesFile,configurationFile )
validator.validate()


from metamorf.engines.engine import Engine
from flask import Flask, Response, json, request
from metamorf.constants import *
from metamorf.tools.manifest import generate_manifest
from metamorf.tools.utils import get_list_nodes_from_metadata
import re


class EndpointDataset(object):

    def __init__(self, manifest):
        self.manifest = manifest

    def __call__(self, **args):
        name_dataset = request.args.get('name')
        if len(args)>0: name_dataset = args['name']
        response = "{}"
        if name_dataset is None: return self.manifest['datasets']
        for pos in self.manifest['datasets']:
            for key, value in pos.items():
                if key.upper() == name_dataset.upper():
                    response = value
        return response


class EndpointNodes(object):

    def __init__(self, manifest):
        self.manifest = manifest

    def __call__(self, **args):
        name_dataset = request.args.get('node')
        if len(args)>0: name_dataset = args['name']
        response = "{}"
        if name_dataset is None: return self.manifest['nodes']
        for pos in self.manifest['nodes']:
                if pos['name'].upper() == name_dataset.upper():
                    response = pos
        return response


class EndpointLineage(object):

    def __init__(self, metadata):
        self.metadata = metadata

    def __call__(self, **args):
        dataset_name = request.args.get('dataset_name')
        column_name = request.args.get('column_name')
        response = dict()
        if len(args)>0:
            dataset_name = args['dataset_name']
            column_name = args['column_name']
        response['name'] = dataset_name
        response['column'] = column_name
        response_origins = []
        if dataset_name is None or column_name is None:
            response['name'] = "Error on INPUT"
            response.pop('column')
            return response
        dataset_spec = self.metadata.get_dataset_spec_from_id_dataset_and_column_name(self.metadata.get_dataset_from_dataset_name(dataset_name).id_dataset,column_name)
        if dataset_spec is None:
            response['column'] = "Column doesn't exist"
            return response

        mapping = self.metadata.get_dataset_t_mapping_from_id_dataset_spec(dataset_spec.id_dataset_spec)
        for map in mapping:
            result = self.replace_mapping(map.value_mapping)
            for r in result:
                new_origin = dict()
                new_origin['branch'] = map.id_branch
                new_origin['origin'] = str(r)

                response_origins.append(new_origin)

        response['source'] = response_origins
        return response

    def replace_mapping(self, mapping):
        all_maps = []
        all_cols = re.findall(r'\[(\d+)\]', mapping)
        for col in all_cols:
            map = self.metadata.get_dataset_t_mapping_from_id_dataset_spec(col)
            if len(map) == 0:
                result = self.metadata.get_tables_dot_column_from_id_dataset_spec(col)
                all_maps.append(mapping.replace("["+col+"]", result))
            else:
                for map_tmp in map:
                    map_tmp = mapping.replace("["+col+"]", map_tmp.value_mapping)
                    map_tmp = self.replace_mapping(map_tmp)
                    all_maps.append(map_tmp)
        return all_maps


class EngineApi(Engine):
    app = None

    def _initialize_engine(self):
        self.engine_name = "Engine API"
        self.engine_command = "api"
        self.app = Flask("Metamorf API")

    def run(self):
        self.start_execution(need_configuration_file=True)
        self.metadata = self.load_metadata(load_om=True, load_entry=False, load_ref=True, load_im=True, owner=self.owner)
        all_nodes = get_list_nodes_from_metadata(self.metadata, self.log, "all")
        self.manifest = generate_manifest(self.metadata, self.configuration_file, all_nodes)

        self.log.log(self.engine_name, "Starting the API Server", LOG_LEVEL_INFO)
        self.add_endpoints()
        self.app.run(host=self.configuration_file['api']['host'], port=self.configuration_file['api']['port'])

        self.finish_execution()

    def add_endpoints(self):
        self.app.add_url_rule('/dataset/<string:name>', 'dataset', EndpointDataset(self.manifest))
        self.app.add_url_rule('/dataset', 'dataset_params', EndpointDataset(self.manifest))
        self.app.add_url_rule('/node/<string:name>', 'node', EndpointNodes(self.manifest))
        self.app.add_url_rule('/node', 'node_params', EndpointNodes(self.manifest))
        self.app.add_url_rule('/lineage/<string:dataset_name>/<string:column_name>', 'lineage', EndpointLineage(self.metadata))
        self.app.add_url_rule('/lineage', 'lineage_params', EndpointLineage(self.metadata))







from metamorf.tools.filecontroller import FileControllerFactory
from metamorf.constants import *
from datetime import datetime
import os

class Log:

    def __init__(self):
        os.system('')
        self.file_log = FileControllerFactory().get_file_reader(FILE_TYPE_LOG)
        self.file_log.set_file_location(ACTUAL_PATH, LOG_FILE_NAME)
        self.file_log.setup_writer(FILE_WRITER_APPEND)

        file_controller_properties = FileControllerFactory().get_file_reader(FILE_TYPE_YML)
        file_controller_properties.set_file_location(os.path.join(PACKAGE_PATH, PROPERTIES_FILE_PATH), PROPERTIES_FILE_NAME)
        self.properties_file = file_controller_properties.read_file()

    def log(self, subject: str, message: str, level: int):
        date_now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        self.file_log.write_file(date_now + " [" + subject + "] " + message + "\n")
        message = self._get_message_with_color(level, message)
        if 'options' not in self.properties_file: return
        if 'log_print' not in self.properties_file['options']: return
        if level == LOG_LEVEL_ONLY_LOG: return
        if self.properties_file['options']['print_debug'] == 'N' and level == LOG_LEVEL_DEBUG: return
        if self.properties_file['options']['log_print'] == "Y":
            print(COLOR_LIGHT_GRAY + date_now + " " + COLOR_DEFAULT + "["
                  + COLOR_DARK_GRAY + subject +
                  COLOR_DEFAULT + "]" + "" + COLOR_DEFAULT + ": " + message)

    def close(self):
        self.file_log.close()

    def _get_message_with_color(self, level: int, message: str):
        result = COLOR_RED + message + COLOR_DEFAULT
        if level == LOG_LEVEL_DEBUG:
            result = COLOR_BLUE + message + COLOR_DEFAULT
        if level == LOG_LEVEL_INFO or level == LOG_LEVEL_ONLY_LOG:
            result = COLOR_DEFAULT + message + COLOR_DEFAULT
        if level == LOG_LEVEL_WARNING:
            result = COLOR_YELLOW + message + COLOR_DEFAULT
        if level == LOG_LEVEL_ERROR:
            result = COLOR_LIGHT_RED + message + COLOR_DEFAULT
        if level == LOG_LEVEL_CRITICAL:
            result = COLOR_RED + message + COLOR_DEFAULT
        if level == LOG_LEVEL_OK:
            result = COLOR_GREEN + message + COLOR_DEFAULT
        return result


from metamorf.engines.engine import Engine
from metamorf.tools.filecontroller import FileControllerFactory
from metamorf.tools.connection import ConnectionFactory
from metamorf.constants import *
import os

class EngineDownload(Engine):

    def _initialize_engine(self):
        self.engine_name = "Engine Download"
        self.engine_command = "download"

    def _initialize_download_engine(self):
        # Default it will download all the files of the main owner
        self.entry_files_to_download = [FILE_ENTRY_AGGREGATORS, FILE_ENTRY_DATASET_MAPPINGS,
                                        FILE_ENTRY_DATASET_RELATIONSHIPS, FILE_ENTRY_ENTITY, FILE_ENTRY_FILTERS,
                                        FILE_ENTRY_ORDER, FILE_ENTRY_PATH, FILE_ENTRY_HAVING,
                                        FILE_ENTRY_DV_ENTITY, FILE_ENTRY_DV_MAPPINGS,FILE_ENTRY_DV_PROPERTIES,
                                        FILE_ENTRY_FILES]
        self.owner_to_download = self.owner

        if 'select' in self.arguments:
            if self.arguments['select'] == "all" or self.arguments['select'] == "*":
                self.entry_files_to_download = [FILE_ENTRY_AGGREGATORS, FILE_ENTRY_DATASET_MAPPINGS,
                                                FILE_ENTRY_DATASET_RELATIONSHIPS, FILE_ENTRY_ENTITY, FILE_ENTRY_FILTERS,
                                                FILE_ENTRY_ORDER, FILE_ENTRY_PATH, FILE_ENTRY_HAVING,
                                                FILE_ENTRY_DV_ENTITY, FILE_ENTRY_DV_MAPPINGS, FILE_ENTRY_DV_PROPERTIES,
                                                FILE_ENTRY_FILES]
            if self.arguments['select'].lower() == FILE_ENTRY_AGGREGATORS.lower():
                self.entry_files_to_download = [FILE_ENTRY_AGGREGATORS]
            if self.arguments['select'].lower() == FILE_ENTRY_DATASET_MAPPINGS.lower():
                self.entry_files_to_download = [FILE_ENTRY_DATASET_MAPPINGS]
            if self.arguments['select'].lower() == FILE_ENTRY_DATASET_RELATIONSHIPS.lower():
                self.entry_files_to_download = [FILE_ENTRY_DATASET_RELATIONSHIPS]
            if self.arguments['select'].lower() == FILE_ENTRY_ENTITY.lower():
                self.entry_files_to_download = [FILE_ENTRY_ENTITY]
            if self.arguments['select'].lower() == FILE_ENTRY_FILTERS.lower():
                self.entry_files_to_download = [FILE_ENTRY_FILTERS]
            if self.arguments['select'].lower() == FILE_ENTRY_ORDER.lower():
                self.entry_files_to_download = [FILE_ENTRY_ORDER]
            if self.arguments['select'].lower() == FILE_ENTRY_PATH.lower():
                self.entry_files_to_download = [FILE_ENTRY_PATH]
            if self.arguments['select'].lower() == FILE_ENTRY_HAVING.lower():
                self.entry_files_to_download = [FILE_ENTRY_HAVING]
            if self.arguments['select'].lower() == FILE_ENTRY_DV_ENTITY.lower():
                self.entry_files_to_download = [FILE_ENTRY_DV_ENTITY]
            if self.arguments['select'].lower() == FILE_ENTRY_DV_MAPPINGS.lower():
                self.entry_files_to_download = [FILE_ENTRY_DV_MAPPINGS]
            if self.arguments['select'].lower() == FILE_ENTRY_DV_PROPERTIES.lower():
                self.entry_files_to_download = [FILE_ENTRY_DV_PROPERTIES]
            if self.arguments['select'].lower() == FILE_ENTRY_FILES.lower():
                self.entry_files_to_download = [FILE_ENTRY_FILES]

        if 'owner' in self.arguments:
            self.owner_to_download = self.arguments['owner']

    def run(self):
        # Starts the execution loading the Configuration File. If there is an error it finishes the execution.
        super().start_execution()
        self._initialize_download_engine()

        self.log.log(self.engine_name, "Start downloading Metadata Entry", LOG_LEVEL_INFO)

        self.metadata_actual = self.load_metadata(load_om=False, load_entry=True, load_ref=False, load_im=False, owner=self.owner_to_download)

        # ENTRY_ENTITY
        content = []
        if FILE_ENTRY_ENTITY in self.entry_files_to_download:
            self.log.log(self.engine_name, "Downloading ["+FILE_ENTRY_ENTITY+"]", LOG_LEVEL_INFO)
            content.append(COLUMNS_ENTRY_ENTITY[:-1])
            for registry in self.metadata_actual.entry_entity:
                content.append(registry.get()[:-1])
            file_controller_final = FileControllerFactory().get_file_reader(FILE_TYPE_CSV)
            file_controller_final.set_file_location(os.path.join(ACTUAL_PATH, ENTRY_FILES_PATH), FILE_ENTRY_ENTITY)
            file_controller_final.setup_writer(FILE_WRITER_NEW_FILE)
            file_controller_final.write_file(content)
            self.log.log(self.engine_name, "Finish download for [" + FILE_ENTRY_ENTITY + "]", LOG_LEVEL_INFO)

        # ENTRY_DATASET_MAPPINGS
        content = []
        if FILE_ENTRY_DATASET_MAPPINGS in self.entry_files_to_download:
            self.log.log(self.engine_name, "Downloading [" + FILE_ENTRY_DATASET_MAPPINGS + "]", LOG_LEVEL_INFO)
            content.append(COLUMNS_ENTRY_DATASET_MAPPINGS[:-1])
            for registry in self.metadata_actual.entry_dataset_mappings:
                content.append(registry.get()[:-1])
            file_controller_final = FileControllerFactory().get_file_reader(FILE_TYPE_CSV)
            file_controller_final.set_file_location(os.path.join(ACTUAL_PATH, ENTRY_FILES_PATH), FILE_ENTRY_DATASET_MAPPINGS)
            file_controller_final.setup_writer(FILE_WRITER_NEW_FILE)
            self.log.log(self.engine_name, "Finish download for [" + FILE_ENTRY_DATASET_MAPPINGS + "]", LOG_LEVEL_INFO)

            file_controller_final.write_file(content)

        # ENTRY_DATASET_RELATIONSHIPS
        content = []
        if FILE_ENTRY_DATASET_RELATIONSHIPS in self.entry_files_to_download:
            self.log.log(self.engine_name, "Downloading [" + FILE_ENTRY_DATASET_RELATIONSHIPS + "]", LOG_LEVEL_INFO)
            content.append(COLUMNS_ENTRY_DATASET_RELATIONSHIPS[:-1])
            for registry in self.metadata_actual.entry_dataset_relationship:
                content.append(registry.get()[:-1])
            file_controller_final = FileControllerFactory().get_file_reader(FILE_TYPE_CSV)
            file_controller_final.set_file_location(os.path.join(ACTUAL_PATH, ENTRY_FILES_PATH), FILE_ENTRY_DATASET_RELATIONSHIPS)
            file_controller_final.setup_writer(FILE_WRITER_NEW_FILE)
            file_controller_final.write_file(content)
            self.log.log(self.engine_name, "Finish download for [" + FILE_ENTRY_DATASET_RELATIONSHIPS + "]", LOG_LEVEL_INFO)

        # ENTRY_ORDER
        content = []
        if FILE_ENTRY_ORDER in self.entry_files_to_download:
            self.log.log(self.engine_name, "Downloading [" + FILE_ENTRY_ORDER + "]", LOG_LEVEL_INFO)
            content.append(COLUMNS_ENTRY_ORDER[:-1])
            for registry in self.metadata_actual.entry_order:
                content.append(registry.get()[:-1])
            file_controller_final = FileControllerFactory().get_file_reader(FILE_TYPE_CSV)
            file_controller_final.set_file_location(os.path.join(ACTUAL_PATH, ENTRY_FILES_PATH), FILE_ENTRY_ORDER)
            file_controller_final.setup_writer(FILE_WRITER_NEW_FILE)
            file_controller_final.write_file(content)
            self.log.log(self.engine_name, "Finish download for [" + FILE_ENTRY_ORDER + "]", LOG_LEVEL_INFO)

        # ENTRY_AGGREGATOR
        content = []
        if FILE_ENTRY_AGGREGATORS in self.entry_files_to_download:
            self.log.log(self.engine_name, "Downloading [" + FILE_ENTRY_AGGREGATORS + "]", LOG_LEVEL_INFO)
            content.append(COLUMNS_ENTRY_AGGREGATORS[:-1])
            for registry in self.metadata_actual.entry_aggregators:
                content.append(registry.get()[:-1])
            file_controller_final = FileControllerFactory().get_file_reader(FILE_TYPE_CSV)
            file_controller_final.set_file_location(os.path.join(ACTUAL_PATH, ENTRY_FILES_PATH), FILE_ENTRY_AGGREGATORS)
            file_controller_final.setup_writer(FILE_WRITER_NEW_FILE)
            file_controller_final.write_file(content)
            self.log.log(self.engine_name, "Finish download for [" + FILE_ENTRY_AGGREGATORS + "]", LOG_LEVEL_INFO)

        # ENTRY_PATH
        content = []
        if FILE_ENTRY_PATH in self.entry_files_to_download:
            self.log.log(self.engine_name, "Downloading [" + FILE_ENTRY_PATH + "]", LOG_LEVEL_INFO)
            content.append(COLUMNS_ENTRY_PATH[:-1])
            for registry in self.metadata_actual.entry_path:
                content.append(registry.get()[:-1])
            file_controller_final = FileControllerFactory().get_file_reader(FILE_TYPE_CSV)
            file_controller_final.set_file_location(os.path.join(ACTUAL_PATH, ENTRY_FILES_PATH), FILE_ENTRY_PATH)
            file_controller_final.setup_writer(FILE_WRITER_NEW_FILE)
            file_controller_final.write_file(content)
            self.log.log(self.engine_name, "Finish download for [" + FILE_ENTRY_PATH + "]", LOG_LEVEL_INFO)

        # ENTRY_FILTERS
        content = []
        if FILE_ENTRY_FILTERS in self.entry_files_to_download:
            self.log.log(self.engine_name, "Downloading [" + FILE_ENTRY_FILTERS + "]", LOG_LEVEL_INFO)
            content.append(COLUMNS_ENTRY_PATH[:-1])
            for registry in self.metadata_actual.entry_filters:
                content.append(registry.get()[:-1])
            file_controller_final = FileControllerFactory().get_file_reader(FILE_TYPE_CSV)
            file_controller_final.set_file_location(os.path.join(ACTUAL_PATH, ENTRY_FILES_PATH), FILE_ENTRY_FILTERS)
            file_controller_final.setup_writer(FILE_WRITER_NEW_FILE)
            file_controller_final.write_file(content)
            self.log.log(self.engine_name, "Finish download for [" + FILE_ENTRY_FILTERS + "]", LOG_LEVEL_INFO)

        # ENTRY_HAVING
        content = []
        if FILE_ENTRY_HAVING in self.entry_files_to_download:
            self.log.log(self.engine_name, "Downloading [" + FILE_ENTRY_HAVING + "]", LOG_LEVEL_INFO)
            content.append(COLUMNS_ENTRY_HAVING[:-1])
            for registry in self.metadata_actual.entry_having:
                content.append(registry.get()[:-1])
            file_controller_final = FileControllerFactory().get_file_reader(FILE_TYPE_CSV)
            file_controller_final.set_file_location(os.path.join(ACTUAL_PATH, ENTRY_FILES_PATH), FILE_ENTRY_HAVING)
            file_controller_final.setup_writer(FILE_WRITER_NEW_FILE)
            file_controller_final.write_file(content)
            self.log.log(self.engine_name, "Finish download for [" + FILE_ENTRY_HAVING + "]", LOG_LEVEL_INFO)

        # ENTRY_DV_ENTITY
        content = []
        if FILE_ENTRY_DV_ENTITY in self.entry_files_to_download:
            self.log.log(self.engine_name, "Downloading [" + FILE_ENTRY_DV_ENTITY + "]", LOG_LEVEL_INFO)
            content.append(COLUMNS_ENTRY_DV_ENTITY[:-1])
            for registry in self.metadata_actual.entry_dv_entity:
                content.append(registry.get()[:-1])
            file_controller_final = FileControllerFactory().get_file_reader(FILE_TYPE_CSV)
            file_controller_final.set_file_location(os.path.join(ACTUAL_PATH, ENTRY_FILES_PATH), FILE_ENTRY_DV_ENTITY)
            file_controller_final.setup_writer(FILE_WRITER_NEW_FILE)
            file_controller_final.write_file(content)
            self.log.log(self.engine_name, "Finish download for [" + FILE_ENTRY_DV_ENTITY + "]", LOG_LEVEL_INFO)

        # ENTRY_DV_MAPPINGS
        content = []
        if FILE_ENTRY_DV_MAPPINGS in self.entry_files_to_download:
            self.log.log(self.engine_name, "Downloading [" + FILE_ENTRY_DV_MAPPINGS + "]", LOG_LEVEL_INFO)
            content.append(COLUMNS_ENTRY_DV_MAPPINGS[:-1])
            for registry in self.metadata_actual.entry_dv_mappings:
                content.append(registry.get()[:-1])
            file_controller_final = FileControllerFactory().get_file_reader(FILE_TYPE_CSV)
            file_controller_final.set_file_location(os.path.join(ACTUAL_PATH, ENTRY_FILES_PATH), FILE_ENTRY_DV_MAPPINGS)
            file_controller_final.setup_writer(FILE_WRITER_NEW_FILE)
            file_controller_final.write_file(content)
            self.log.log(self.engine_name, "Finish download for [" + FILE_ENTRY_DV_MAPPINGS + "]", LOG_LEVEL_INFO)

        # ENTRY_DV_PROPERTIES
        content = []
        if FILE_ENTRY_DV_PROPERTIES in self.entry_files_to_download:
            self.log.log(self.engine_name, "Downloading [" + FILE_ENTRY_DV_PROPERTIES + "]", LOG_LEVEL_INFO)
            content.append(COLUMNS_ENTRY_DV_PROPERTIES[:-1])
            for registry in self.metadata_actual.entry_dv_properties:
                content.append(registry.get()[:-1])
            file_controller_final = FileControllerFactory().get_file_reader(FILE_TYPE_CSV)
            file_controller_final.set_file_location(os.path.join(ACTUAL_PATH, ENTRY_FILES_PATH), FILE_ENTRY_DV_PROPERTIES)
            file_controller_final.setup_writer(FILE_WRITER_NEW_FILE)
            file_controller_final.write_file(content)
            self.log.log(self.engine_name, "Finish download for [" + FILE_ENTRY_DV_PROPERTIES + "]", LOG_LEVEL_INFO)

        # ENTRY_FILES
        content = []
        if FILE_ENTRY_FILES in self.entry_files_to_download:
            self.log.log(self.engine_name, "Downloading [" + FILE_ENTRY_FILES + "]", LOG_LEVEL_INFO)
            content.append(COLUMNS_ENTRY_FILES[:-1])
            for registry in self.metadata_actual.entry_files:
                content.append(registry.get()[:-1])
            file_controller_final = FileControllerFactory().get_file_reader(FILE_TYPE_CSV)
            file_controller_final.set_file_location(os.path.join(ACTUAL_PATH, ENTRY_FILES_PATH), FILE_ENTRY_FILES)
            file_controller_final.setup_writer(FILE_WRITER_NEW_FILE)
            file_controller_final.write_file(content)
            self.log.log(self.engine_name, "Finish download for [" + FILE_ENTRY_FILES + "]", LOG_LEVEL_INFO)



        super().finish_execution()


import sys, getopt
from metamorf.tools.log import Log
from metamorf.constants import *
from metamorf.engines.engine_init import EngineInit
from metamorf.engines.engine_deploy import EngineDeploy
from metamorf.engines.engine_validate import EngineValidate
from metamorf.engines.engine_upload import EngineUpload
from metamorf.engines.engine_process import EngineProcess
from metamorf.engines.engine_deploy_example import EngineDeployExample
from metamorf.engines.engine_download import EngineDownload
from metamorf.engines.engine_recover import EngineRecover
from metamorf.engines.engine_delete import EngineDelete
from metamorf.engines.engine_output import EngineOutput
from metamorf.engines.engine_run import EngineRun
from metamorf.engines.engine_backup import EngineBackup
from metamorf.engines.engine_manifest import EngineManifest
from metamorf.engines.engine_api import EngineApi
from metamorf.engines.engine_restore import EngineRestore
from metamorf.engines.engine_files import EngineFiles
from metamorf.engines.engine_metadata import EngineMetadata
from metamorf.tools.argparser import ArgParser
from metamorf.engines.engine_commit import EngineCommit
from metamorf.help import print_help

def main():
    os.system('')
    argp = ArgParser(sys.argv[1:])
    arguments = argp.get_arguments_parsed()
    if 'error' in arguments:
        print(COLOR_LIGHT_RED + "Error executing Metamorf: " + arguments['error'] + "\nFor more information run [metamorf help]" + COLOR_LIGHT_GRAY)
        return
    if arguments['command'] == 'help':
        print_help()
        return

    # SETUP LOG
    log = Log()
    log.log('METAMORF', '*********************************************************************', LOG_LEVEL_INFO)
    log.log('METAMORF', 'Running metamorf version "' + VERSION + '"', LOG_LEVEL_INFO)

    # COMMANDS
    if arguments['command'] == 'init':
        engine = EngineInit(log, arguments)
        engine.run()
    elif arguments['command'] == 'deploy':
        engine = EngineDeploy(log, arguments)
        engine.run()
    elif arguments['command'] == 'deploy-example':
        engine = EngineDeployExample(log, arguments)
        engine.run()
    elif arguments['command'] == 'validate':
        engine = EngineValidate(log, arguments)
        engine.run()
    elif arguments['command'] == 'upload':
        engine = EngineUpload(log, arguments)
        engine.run()
    elif arguments['command'] == 'process':
        engine = EngineProcess(log, arguments)
        engine.run()
    elif arguments['command'] == 'download':
        engine = EngineDownload(log, arguments)
        engine.run()
    elif arguments['command'] == 'commit':
        engine = EngineCommit(log, arguments)
        engine.run()
    elif arguments['command'] == 'recover':
        engine = EngineRecover(log, arguments)
        engine.run()
    elif arguments['command'] == 'delete':
        engine = EngineDelete(log, arguments)
        engine.run()
    elif arguments['command'] == 'output':
        engine = EngineOutput(log, arguments)
        engine.run()
    elif arguments['command'] == 'run':
        engine = EngineRun(log, arguments)
        engine.run()
    elif arguments['command'] == 'backup':
        engine = EngineBackup(log, arguments)
        engine.run()
    elif arguments['command'] == 'api':
        engine = EngineApi(log, arguments)
        engine.run()
    elif arguments['command'] == 'manifest':
        engine = EngineManifest(log, arguments)
        engine.run()
    elif arguments['command'] == 'restore':
        engine = EngineRestore(log, arguments)
        engine.run()
    elif arguments['command'] == 'files':
        engine = EngineFiles(log, arguments)
        engine.run()
    elif arguments['command'] == 'metadata':
        engine = EngineMetadata(log, arguments)
        engine.run()
    else:
        log.log('METAMORF', 'Command under development. More information at https://www.metamorf.io', LOG_LEVEL_CRITICAL)

    log.close()

if __name__ == "__main__":
   main()



from metamorf.tools.connection import ConnectionFactory, Connection
from metamorf.constants import *
from metamorf.tools.filecontroller import *
from metamorf.tools.log import *
from metamorf.tools.database_objects import *


class Test:

    def run(self):
        file_controller_configuration = FileControllerFactory().get_file_reader(FILE_TYPE_YML)
        file_controller_configuration.set_file_location(ACTUAL_PATH, CONFIGURATION_FILE_NAME)
        self.configuration_file = file_controller_configuration.read_file()

        log = Log()
        self.connection = ConnectionFactory().get_connection(CONNECTION_SQLITE)
        self.connection.setup_connection(self.configuration_file['metadata'], log)

        query_create_table = "CREATE TABLE EJEMPLO(ID integer, NAME text, FECHA date)"
        #self.connection.execute(query_create_table)

        query_insert_into = "INSERT INTO ejemplo(id,name,fecha) values(?,?,?)"
        values = (1,'Aumatell, Guillermo', '2023-01-29')

        query_insert_into_total = "INSERT INTO ejemplo(id,name,fecha) values(2,'Arbona, Emma','2023-01-29')"
        self.connection.execute(query_insert_into_total)
        rows = self.connection.get_query_result()
        print(rows)

        select_query = "SELECT ID,NAME,FECHA FROM EJEMPLO"
        self.connection.execute(select_query)
        rows = self.connection.get_query_result()
        print(rows)
        for r in rows:
            print(type(r))

        #self.connection.cursor.execute(query_insert_into, values)
        #self.connection.execute(query_insert_into_total)
        #self.connection.commit()



        log.close()

#test = Test().run()
print(False and True)
print(False and False)
print(True and True)
print(False and False)

'''
om_eo = EntryOrder('EJEMPLO', 'ID', 'desc')
        print(om_eo)
        om_ea = EntryAggregators('A','B', 'ID',2)
        print(om_ea)
        om_ef = EntryFilters('A','a=1',1)
        print(om_ef)
        om_er = EntryDatasetRelationships('A','a','B','b', 'master join')
        print(om_er)
        om_ee = EntryEntity('entity' ,'A', 'TMP', 'path')
        print(om_ee)
        om_ep = EntryPath('COD_PATH','DATABASE', 'SCHEMA')
        print(om_ep)
        om_ep = EntryPath('COD_PATH', '', '')
        print(om_ep)
        om_ep = EntryPath(cod_path='COD_PATH', database_name=None, schema_name=None)
        print(om_ep)
        om_edm = EntryDatasetMappings('ENTITY_SOURCE','SUBSTR(CAMPO_A)','ENTITY_TARGET', 'CAMPO_A_GUAY','VARCHAR', 1,'PK',0)
        print(om_edm)
        om_edm = EntryDatasetMappings('ENTITY_SOURCE', 'SUBSTR(CAMPO_A)', 'ENTITY_TARGET', 'CAMPO_A_GUAY', 'VARCHAR', 1, None, 0)
        print(om_edm)
        '''


from abc import ABC, abstractmethod
import sqlite3
import snowflake.connector
import mysql.connector
from metamorf.constants import *
import hashlib
import fdb
import psycopg2
from metamorf.tools.log import Log
from metamorf.tools.database_objects import Column

class Connection(ABC):

    def __init__(self):
        self.conn = None
        self.is_executing = False
        self.engine_name = "Connection"

    def setup_connection(self, configuration: dict, log: Log):
        self.log = log
        self.get_connection(configuration)
        self._get_cursor()
        self.setup_after_connection_established()

    @abstractmethod
    def get_connection(self, configuration: dict):
        pass

    @abstractmethod
    def get_connection_type(self):
        pass

    @abstractmethod
    def get_sysdate_value(self):
        pass

    @abstractmethod
    def get_sysdate_value_infinite(self):
        pass

    @abstractmethod
    def get_table_columns_definition(self, table_name):
        pass

    @abstractmethod
    def does_table_exists(self, table_name):
        pass

    def setup_after_connection_established(self):
        pass

    def _get_cursor(self):
        if self.conn is None:
            self.cursor = None
        else:
            self.cursor = self.conn.cursor()

    def commit(self):
        self.conn.commit()

    def rollback(self):
        self.conn.rollback()

    def close(self):
        try:
            self.conn.close()
        except:
            pass

    def execute(self, all_queries: str):
        """Return a boolean with the result of the execution."""
        self.is_executing = True
        if self.cursor is None:
            self.log.log(self.engine_name, 'The connection is not prepared to execute', LOG_LEVEL_ERROR)
            self.is_executing = False
            return False


        if len(all_queries) == 0:
            self.log.log(self.engine_name, 'There\'s nothing to load', LOG_LEVEL_ONLY_LOG)
            return True
        queries = all_queries.split(";")
        num_queries = len(queries)
        if queries[-1] == "": num_queries -= 1
        if num_queries==1: self.log.log(self.engine_name, "Starting to execute " + str(num_queries) + " query", LOG_LEVEL_ONLY_LOG)
        if num_queries > 1: self.log.log(self.engine_name, "Starting to execute " + str(num_queries) + " queries", LOG_LEVEL_ONLY_LOG)

        num_queries_executed = 0
        for q in queries:
            if q is None or q.strip() == '':
                continue
            try:
                q = q.strip()
                self.cursor.execute(q)
                self.log.log(self.engine_name, "Query Executed:\n" + q, LOG_LEVEL_ONLY_LOG)
                num_queries_executed += 1
            except Exception as e:
                self.log.log(self.engine_name, "Query Executed:\n" + q, LOG_LEVEL_ONLY_LOG)
                self.log.log(self.engine_name, "Error: " + str(e), LOG_LEVEL_ERROR)
                self.is_executing = False
                return False
            #if(self.connection_type == CONNECTION_FIREBIRD): self.connection.commit()
        if num_queries == 1: self.log.log(self.engine_name, "Query Finished Ok", LOG_LEVEL_ONLY_LOG)
        if num_queries > 1: self.log.log(self.engine_name, "All queries finished Ok", LOG_LEVEL_ONLY_LOG)
        self.is_executing = False
        return True

    def get_query_result(self):
        return self.cursor.fetchall()

    def get_md5_function(self):
        return "MD5([x])"

    @abstractmethod
    def get_hash_for_metadata(self):
        pass

    @abstractmethod
    def get_sysdate_for_metadata(self):
        pass

    @abstractmethod
    def get_cast_string_for_metadata(self):
        pass

    @abstractmethod
    def get_integer_for_metadata(self):
        pass

    @abstractmethod
    def get_string_for_metadata(self):
        pass


class ConnectionSQLite(Connection):
    def get_connection_type(self):
        return CONNECTION_SQLITE

    def get_connection(self, configuration: dict):
        self.log.log(self.engine_name, "Trying to connect to the SQLite database", LOG_LEVEL_ONLY_LOG)
        self.conn = sqlite3.connect(configuration['sqlite_path'])
        sqlite3.enable_callback_tracebacks(True)
        self.log.log(self.engine_name, "Connection to the SQLite database established", LOG_LEVEL_ONLY_LOG)

    # Usage on the SQLITE connection because hashing functions doesn't exist on this db
    def _md5(self,word):
        final_word = str(word)
        return hashlib.md5(final_word.encode('utf-8')).hexdigest()

    def setup_after_connection_established(self):
        self.log.log(self.engine_name, "Adding MD5 Function to the SQLite Database", LOG_LEVEL_ONLY_LOG)
        self.conn.create_function("MD5", 1, self._md5)
        self.log.log(self.engine_name, "Added MD5 Function to the SQLite Database", LOG_LEVEL_ONLY_LOG)

    def get_sysdate_value(self):
        return 'datetime(\'now\')'

    def get_sysdate_value_infinite(self):
        return "DATETIME(\'9999-12-31 00:00:00\')"

    def get_table_columns_definition(self, table_name):
        query_to_execute = "PRAGMA TABLE_INFO(" + table_name + ")"
        self.execute(query_to_execute)
        result = self.get_query_result()
        all_columns = []
        for r in result:
            if r[3] == 1: is_nullable = 1
            else: is_nullable = 0
            pk_result = "NULL"
            if r[5] == 1: pk_result="PK"
            all_columns.append(Column(r[0], r[1], r[2], r[4], pk_result, is_nullable,0,0,0))
        return all_columns

    def does_table_exists(self, table_name):
        query_to_execute = "PRAGMA TABLE_INFO(" + table_name + ")"
        self.execute(query_to_execute)
        result = self.get_query_result()
        if len(result)>0: return True
        return False

    def get_hash_for_metadata(self):
        return ['TEXT', 0, 0]

    def get_sysdate_for_metadata(self):
        return ['TEXT', 0, 0]

    def get_cast_string_for_metadata(self):
        return 'CAST([x] as text)'

    def get_integer_for_metadata(self):
        return ['INTEGER', 0, 0]

    def get_string_for_metadata(self):
        return ['TEXT', 0, 0]

class ConnectionMySQL(Connection):
    def get_connection_type(self):
        return CONNECTION_MYSQL

    def get_connection(self, configuration: dict):
        self.log.log(self.engine_name, "Trying to connect to the MySQL database", LOG_LEVEL_ONLY_LOG)

        self.conn = mysql.connector.connect(
            user = configuration['mysql_user'],
            password = configuration['mysql_password'],
            host = configuration['mysql_host'],
            database = configuration['mysql_database']
        )
        self.log.log(self.engine_name, "Finished to connect to the MySQL database", LOG_LEVEL_ONLY_LOG)

    def setup_after_connection_established(self):
        self.activate_concate_with_pipes()

    def activate_concate_with_pipes(self):
        self.log.log(self.engine_name, "Activating concat by pipes", LOG_LEVEL_ONLY_LOG)
        set_on_concat_by_pipes = "SET sql_mode=(SELECT CONCAT(@@sql_mode,',PIPES_AS_CONCAT')) "
        self.execute(set_on_concat_by_pipes)
        self.log.log(self.engine_name, "Activation finished", LOG_LEVEL_ONLY_LOG)

    def get_sysdate_value(self):
        return 'CURRENT_TIMESTAMP()'

    def get_sysdate_value_infinite(self):
        return "TIMESTAMP(\'9999-12-31 00:00:00\')"

    def get_table_columns_definition(self, table_name):
        # ID, COLUMN NAME, TYPE, DEFAULT_VALUE, PK 0/1, IS_NULLABLE, LENGTH, PRECISION, SCALE
        query_to_execute =  "select ordinal_position, column_name, data_type, COLUMN_DEFAULT, COLUMN_KEY, is_nullable, character_maximum_length, numeric_precision, numeric_scale " +\
                            " FROM information_schema.columns " +\
                            " WHERE table_name = '"+table_name+"' "
        self.execute(query_to_execute)
        result = self.get_query_result()
        all_columns = []
        for r in result:
            if r[5] == 'YES':
                is_nullable = 1
            else:
                is_nullable = 0
            if r[6] is None: length = 0
            else: length = r[6]
            if r[7] is None: precision = 0
            else: precision = r[7]
            if r[8] is None: scale = 0
            else: scale = r[8]
            pk_result = "NULL"
            if r[4] == "PRI": pk_result="PK"
            #                         id, name, type, default, is_pk, is_nullable, length, precision, scale
            all_columns.append(Column(r[0], r[1], r[2], r[3], pk_result, is_nullable, length, precision, scale))
        return all_columns

    def does_table_exists(self, table_name):
        query_to_execute = "select column_name FROM information_schema.columns WHERE table_name = '" + table_name + "'"
        self.execute(query_to_execute)
        result = self.get_query_result()
        if len(result) > 0: return True
        return False

    def get_hash_for_metadata(self):
        return ['BINARY', 32, 0]

    def get_sysdate_for_metadata(self):
        return ['DATETIME', 0, 0]

    def get_cast_string_for_metadata(self):
        return 'CAST([x] as CHAR)'

    def get_integer_for_metadata(self):
        return ['INTEGER', 0, 0]

    def get_string_for_metadata(self):
        return ['VARCHAR', 0, 0]


class ConnectionPostgreSQL(Connection):
    def get_connection_type(self):
        return CONNECTION_POSTGRESQL

    def get_connection(self, configuration: dict):
        self.log.log(self.engine_name, "Trying to connect to the PostgreSQL database", LOG_LEVEL_ONLY_LOG)
        self.conn = psycopg2.connect(
            user = configuration['postgres_user'],
            password = configuration['postgres_password'],
            host = configuration['postgres_host'],
            database = configuration['postgres_database'],
            options= '-c search_path='+configuration['postgres_schema']
        )
        self.log.log(self.engine_name, "Finished to connect to the PostgreSQL database", LOG_LEVEL_ONLY_LOG)

    def get_sysdate_value(self):
        return 'CURRENT_TIMESTAMP'

    def get_sysdate_value_infinite(self):
        return "TO_TIMESTAMP(\'9999-12-31 00:00:00\', \'YYYY-MM-DD HH24:MI:SS\')"

    def get_table_columns_definition(self, table_name):
        table_name = table_name.lower()
        # ID, COLUMN NAME, TYPE, DEFAULT_VALUE, PK 0/1, IS_NULLABLE, LENGTH, PRECISION, SCALE
        query_to_execute =  "SELECT  ordinal_position, column_name, udt_name, column_default, 0 as is_pk, is_nullable, character_maximum_length , numeric_precision , numeric_scale " +\
                            "FROM information_schema.columns " +\
                            "WHERE  table_name = '" + table_name + "'"
        self.execute(query_to_execute)
        result = self.get_query_result()
        all_columns = []
        for r in result:
            if r[5] == 'YES':
                is_nullable = 1
            else:
                is_nullable = 0
            if r[6] is None: length = 0
            else: length = r[6]
            if r[7] is None: precision = 0
            else: precision = r[7]
            if r[8] is None: scale = 0
            else: scale = r[8]
            pk_result = "NULL"
            if r[4] == "PRI": pk_result="PK"
            #                         id, name, type, default, is_pk, is_nullable, length, precision, scale
            all_columns.append(Column(r[0], r[1].upper(), r[2].upper(), r[3], pk_result, is_nullable, length, precision, scale))
        return all_columns

    def does_table_exists(self, table_name):
        table_name = table_name.lower()
        query_to_execute = "select column_name FROM information_schema.columns WHERE table_name = '" + table_name + "'"
        self.execute(query_to_execute)
        result = self.get_query_result()
        if len(result) > 0: return True
        return False

    def get_hash_for_metadata(self):
        return ['TEXT', 0, 0]

    def get_sysdate_for_metadata(self):
        return ['TIMESTAMP', 0, 0]

    def get_cast_string_for_metadata(self):
        return 'CAST([x] as text)'

    def get_integer_for_metadata(self):
        return ['INTEGER', 0, 0]

    def get_string_for_metadata(self):
        return ['VARCHAR', 0, 0]


class ConnectionFirebird(Connection):
    def get_connection_type(self):
        return CONNECTION_FIREBIRD

    def get_connection(self, configuration: dict):
        self.conn = fdb.connect(
            host=configuration['firebird_host'],
            database=configuration['firebird_database'],
            user=configuration['firebird_user'],
            password=configuration['firebird_password']
        )


class ConnectionSnowflake(Connection):
    def get_connection_type(self):
        return CONNECTION_SNOWFLAKE

    def get_connection(self, configuration: dict):
        self.log.log(self.engine_name, "Trying to connect to the Snowflake database", LOG_LEVEL_ONLY_LOG)
        self.schema_connection = configuration['snowflake_schema']
        try:
            self.conn = snowflake.connector.connect(
                user = configuration['snowflake_user'],
                password = configuration['snowflake_password'],
                account = configuration['snowflake_account'],
                warehouse = configuration['snowflake_warehouse'],
                database = configuration['snowflake_database'],
                schema = configuration['snowflake_schema'],
                role = configuration['snowflake_role'],
                autocommit = False
            )
            self.log.log(self.engine_name, "Trying to connect the Snowflake database finished", LOG_LEVEL_ONLY_LOG)
        except Exception as e:
            self.log.log(self.engine_name, "Trying to connect the Snowflake database finished", LOG_LEVEL_ONLY_LOG)
            self.log.log(self.engine_name, str(e), LOG_LEVEL_ERROR)

    def get_table_columns_definition(self, table_name):
        # ID, COLUMN NAME, TYPE, DEFAULT_VALUE, PK 0/1, IS_NULLABLE, LENGTH, PRECISION, SCALE
        query_to_execute = "select ORDINAL_POSITION, COLUMN_NAME, DATA_TYPE, IS_NULLABLE, COLUMN_DEFAULT, 0 as IS_PK, "+\
                           " CHARACTER_MAXIMUM_LENGTH, NUMERIC_PRECISION, NUMERIC_SCALE "+\
                           " from INFORMATION_SCHEMA.COLUMNS where TABLE_NAME='"+table_name+"' and TABLE_SCHEMA = '"+self.schema_connection+"'"
        self.execute(query_to_execute)
        result = self.get_query_result()
        all_columns = []
        for r in result:
            if r[3]=='YES': is_nullable = 1
            else: is_nullable = 0
            if r[6] is None: length = 0
            else: length = r[6]
            if r[7] is None: precision = 0
            else: precision = r[7]
            if r[8] is None: scale = 0
            else: scale = r[8]
            pk_result = "NULL"
            if r[5] == 1: pk_result="PK"
            #                         id, name, type, default, is_pk, is_nullable, length, precision, scale
            all_columns.append(Column(r[0], r[1], r[2], r[4], pk_result, is_nullable, length, precision, scale))
        return all_columns

    def does_table_exists(self, table_name):
        query_to_execute = "SELECT * FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_NAME='" + table_name + "'"
        self.execute(query_to_execute)
        result = self.get_query_result()
        if len(result) > 0: return True
        return False

    def get_sysdate_value(self):
        return 'CURRENT_TIMESTAMP()'

    def get_sysdate_value_infinite(self):
        return "TO_TIMESTAMP(\'9999-12-31 00:00:00\', \'YYYY-MM-DD HH24:MI:SS\')"

    def get_hash_for_metadata(self):
        return ['VARCHAR', 0, 0]

    def get_sysdate_for_metadata(self):
        return ['TIMESTAMP', 0, 0]

    def get_cast_string_for_metadata(self):
        return 'TO_CHAR([x])'

    def get_integer_for_metadata(self):
        return ['NUMBER', 0, 0]

    def get_string_for_metadata(self):
        return ['VARCHAR', 0, 0]


#####################################################################################

class ConnectionFactory:
    def get_connection(self, name_connection: str):
        if name_connection.upper() == CONNECTION_SNOWFLAKE.upper():
            return ConnectionSnowflake()
        elif name_connection.upper() == CONNECTION_SQLITE.upper():
            return ConnectionSQLite()
        elif name_connection.upper() == CONNECTION_FIREBIRD.upper():
            return ConnectionFirebird()
        elif name_connection.upper() == CONNECTION_MYSQL.upper():
            return ConnectionMySQL()
        elif name_connection.upper() == CONNECTION_POSTGRESQL.upper():
            return ConnectionPostgreSQL()
        else:
            return None




from metamorf.engines.engine import Engine
from metamorf.tools.filecontroller import FileControllerFactory
from metamorf.tools.connection import ConnectionFactory
from metamorf.constants import *

class EngineValidate(Engine):

    def _initialize_engine(self):
        self.engine_name = "Engine Validate"
        self.engine_command = "validate"

    def run(self):
        # Starts the execution loading the Configuration File. If there is an error it finishes the execution.
        super().start_execution()

        self.log.log(self.engine_name, "Starting to validate metadata connection", LOG_LEVEL_INFO)
        connection_type = self.configuration_file['metadata']['connection_type']
        connection = ConnectionFactory().get_connection(connection_type)
        connection.setup_connection(self.configuration_file['metadata'], self.log)
        self.log.log(self.engine_name, "Metadata connection validation finished", LOG_LEVEL_INFO)

        self.log.log(self.engine_name, "Starting to validate data connection", LOG_LEVEL_INFO)
        connection_type = self.configuration_file['data']['connection_type']
        connection = ConnectionFactory().get_connection(connection_type)
        connection.setup_connection(self.configuration_file['data'], self.log)
        self.log.log(self.engine_name, "Data connection validation finished", LOG_LEVEL_INFO)


        super().finish_execution()


from metamorf.tools.query import Query
from metamorf.constants import *
import time

class Node:

    def __init__(self, name: str, query: Query, predecessors: list[str]):
        '''Name: Name of the process / table that will be loaded
        Query: Process to be executed, need to set the QUERY_TYPE that will be generated
        Predecessors: list string of the predecessors Nodes.name
        '''
        self.query = query
        self.predecessors = predecessors
        self.name = name
        self.status = NODE_STATUS_WAITING
        self.time = 0

    def set_node_type(self, node_type: int):
        self.query.set_type(node_type)

    def init_execution(self):
        self.time = float(time.time())

    def finish_execution(self):
        self.time = float(time.time()) - float(self.time)



import os

# METAMORF PROPERTIES
VERSION = "0.4.4.1"

# Constants for QUERY TYPES
QUERY_TYPE_INSERT = 0
QUERY_TYPE_UPDATE = 1
QUERY_TYPE_VIEW = 2
QUERY_TYPE_DELETE = 3
QUERY_TYPE_SELECT = 4
QUERY_TYPE_MERGE = 5
QUERY_TYPE_TRUNCATE_AND_INSERT = 6
# Query types for the engines
QUERY_TYPE_VALUES = 99


# Constants for JOIN TYPES
JOIN_TYPE_INNER = 0
JOIN_TYPE_MASTER = 1
JOIN_TYPE_DETAIL = 2
JOIN_TYPE_OUTER = 3

# Constants for FILES TYPES EXTENSIONS
FILE_TYPE_YML = "yml"
FILE_TYPE_SQL = "sql"
FILE_TYPE_LOG = "log"
FILE_TYPE_TXT = "txt"
FILE_TYPE_CSV = "csv"
FILE_TYPE_JSON = "json"

# Constants for CONNECTION TYPES
CONNECTION_SQLITE = "SQLITE"
CONNECTION_MYSQL = "MYSQL"
CONNECTION_SNOWFLAKE = "SNOWFLAKE"
CONNECTION_FIREBIRD = "FIREBIRD"
CONNECTION_POSTGRESQL = "POSTGRESQL"

CONNECTIONS_WITHOUT_OUTER_JOIN = [CONNECTION_SQLITE, CONNECTION_MYSQL]

# Constants for ENGINES
MODULE_ELT = "ELT"
MODULE_DV = "DATAVAULT"

# Constants for Configuration Location
LOG_FILE_NAME = "metamorf"
CONFIGURATION_FILE_PATH = "configuration"
CONFIGURATION_FILE_NAME = 'configuration.yml'
CONFIGURATION_FILE_DATA = 'configuration_data_'
CONFIGURATION_FILE_METADATA = 'configuration_metadata_'
PROPERTIES_FILE_PATH = "configuration"
PROPERTIES_FILE_NAME = 'properties.yml'
INITIALIZATION_FILE_PATH = "initialization"
INITIALIZATION_FILE_NAME = 'init_'
ENTRY_FILES_PATH = "entry"
OUTPUT_FILES_PATH = 'output'
BACKUP_FILES_PATH = 'backup'
ACTUAL_PATH = os.getcwd()
PACKAGE_PATH = os.path.dirname(__file__)

# FILE WRITER
FILE_WRITER_NEW_FILE = "w"
FILE_WRITER_APPEND = "a"

# LOG System
LOG_LEVEL_ONLY_LOG = 7
LOG_LEVEL_CRITICAL = 6
LOG_LEVEL_ERROR = 5
LOG_LEVEL_WARNING = 4
LOG_LEVEL_OK = 3
LOG_LEVEL_INFO = 2
LOG_LEVEL_DEBUG = 1

# Colors
COLOR_DEFAULT = "\033[0m"
COLOR_BLACK = "\033[0;30m"
COLOR_RED = "\033[0;31m"
COLOR_GREEN = "\033[0;32m"
COLOR_BROWN = "\033[0;33m"
COLOR_BLUE = "\033[0;34m"
COLOR_PURPLE = "\033[0;35m"
COLOR_CYAN = "\033[0;36m"
COLOR_LIGHT_GRAY = "\033[0;37m"
COLOR_DARK_GRAY = "\033[1;30m"
COLOR_LIGHT_RED = "\033[1;31m"
COLOR_LIGHT_GREEN = "\033[1;32m"
COLOR_YELLOW = "\033[1;33m"
COLOR_LIGHT_BLUE = "\033[1;34m"
COLOR_LIGHT_PURPLE = "\033[1;35m"
COLOR_LIGHT_CYAN = "\033[1;36m"
COLOR_LIGHT_WHITE = "\033[1;37m"

# ENTRY FILES
FILE_ENTRY_AGGREGATORS = 'ENTRY_AGGREGATORS'
FILE_ENTRY_DATASET_MAPPINGS = 'ENTRY_DATASET_MAPPINGS'
FILE_ENTRY_DATASET_RELATIONSHIPS = 'ENTRY_DATASET_RELATIONSHIPS'
FILE_ENTRY_ENTITY = 'ENTRY_ENTITY'
FILE_ENTRY_FILTERS = 'ENTRY_FILTERS'
FILE_ENTRY_ORDER = 'ENTRY_ORDER'
FILE_ENTRY_PATH = 'ENTRY_PATH'
FILE_ENTRY_HAVING = 'ENTRY_HAVING'
FILE_ENTRY_DV_ENTITY = 'ENTRY_DV_ENTITY'
FILE_ENTRY_DV_MAPPINGS = 'ENTRY_DV_MAPPINGS'
FILE_ENTRY_DV_PROPERTIES = 'ENTRY_DV_PROPERTIES'
FILE_ENTRY_FILES = 'ENTRY_FILES'

# MANIFEST FILE
FILE_MANIFEST = 'manifest'

# TABLES
TABLE_ENTRY_AGGREGATORS = 'ENTRY_AGGREGATORS'
TABLE_ENTRY_DATASET_MAPPINGS = 'ENTRY_DATASET_MAPPINGS'
TABLE_ENTRY_DATASET_RELATIONSHIPS = 'ENTRY_DATASET_RELATIONSHIPS'
TABLE_ENTRY_ENTITY = 'ENTRY_ENTITY'
TABLE_ENTRY_FILTERS = 'ENTRY_FILTERS'
TABLE_ENTRY_HAVING = 'ENTRY_HAVING'
TABLE_ENTRY_ORDER = 'ENTRY_ORDER'
TABLE_ENTRY_PATH = 'ENTRY_PATH'
TABLE_ENTRY_DV_MAPPINGS = 'ENTRY_DV_MAPPINGS'
TABLE_ENTRY_DV_ENTITY = 'ENTRY_DV_ENTITY'
TABLE_ENTRY_DV_PROPERTIES = 'ENTRY_DV_PROPERTIES'
TABLE_ENTRY_FILES = 'ENTRY_FILES'
TABLE_OM_DATASET = 'OM_DATASET'
TABLE_OM_DATASET_EXECUTION = 'OM_DATASET_EXECUTION'
TABLE_OM_DATASET_T_ORDER = 'OM_DATASET_T_ORDER'
TABLE_OM_DATASET_T_AGG = 'OM_DATASET_T_AGG'
TABLE_OM_DATASET_T_DISTINCT = 'OM_DATASET_T_DISTINCT'
TABLE_OM_DATASET_SPECIFICATION = 'OM_DATASET_SPECIFICATION'
TABLE_OM_DATASET_RELATIONSHIPS = 'OM_DATASET_RELATIONSHIPS'
TABLE_OM_DATASET_T_MAPPING = 'OM_DATASET_T_MAPPING'
TABLE_OM_DATASET_T_FILTER = 'OM_DATASET_T_FILTER'
TABLE_OM_DATASET_T_HAVING = 'OM_DATASET_T_HAVING'
TABLE_OM_DATASET_PATH = 'OM_DATASET_PATH'
TABLE_OM_PROPERTIES = 'OM_PROPERTIES'
TABLE_OM_DATASET_HARDCODED = 'OM_DATASET_HARDCODED'
TABLE_OM_DATASET_DV = 'OM_DATASET_DV'
TABLE_OM_DATASET_FILE = 'OM_DATASET_FILE'
TABLE_OM_REF_QUERY_TYPE = 'OM_REF_QUERY_TYPE'
TABLE_OM_REF_KEY_TYPE = 'OM_REF_KEY_TYPE'
TABLE_OM_REF_ENTITY_TYPE = 'OM_REF_ENTITY_TYPE'
TABLE_OM_REF_JOIN_TYPE = 'OM_REF_JOIN_TYPE'
TABLE_OM_REF_MODULES = 'OM_REF_MODULES'
TABLE_OM_REF_ORDER_TYPE = 'OM_REF_ORDER_TYPE'

# TABLES GIT
TABLE_GIT_ENTRY_AGGREGATORS = 'GIT_ENTRY_AGGREGATORS'
TABLE_GIT_ENTRY_DATASET_MAPPINGS = 'GIT_ENTRY_DATASET_MAPPINGS'
TABLE_GIT_ENTRY_DATASET_RELATIONSHIPS = 'GIT_ENTRY_DATASET_RELATIONSHIPS'
TABLE_GIT_ENTRY_ENTITY = 'GIT_ENTRY_ENTITY'
TABLE_GIT_ENTRY_FILTERS = 'GIT_ENTRY_FILTERS'
TABLE_GIT_ENTRY_HAVING = 'GIT_ENTRY_HAVING'
TABLE_GIT_ENTRY_ORDER = 'GIT_ENTRY_ORDER'
TABLE_GIT_ENTRY_PATH = 'GIT_ENTRY_PATH'
TABLE_GIT_ENTRY_DV_ENTITY = 'GIT_ENTRY_DV_ENTITY'
TABLE_GIT_ENTRY_DV_MAPPINGS = 'GIT_ENTRY_DV_MAPPINGS'
TABLE_GIT_ENTRY_DV_PROPERTIES = 'GIT_ENTRY_DV_PROPERTIES'
TABLE_GIT_ENTRY_FILES = 'GIT_ENTRY_FILES'

# TABLES IM
TABLE_IM_OM_DATASET_INFORMATION = "OM_DATASET_INFORMATION"
TABLE_IM_OM_RELATIONSHIPS = "OM_RELATIONSHIPS"
TABLE_IM_OM_DATASET_SPECIFICATION_INFORMATION = "OM_DATASET_SPECIFICATION_INFORMATION"

# COLUMNS METADATA
## Entry Metadata
COLUMNS_ENTRY_AGGREGATORS = ['COD_ENTITY_TARGET','COD_ENTITY_SOURCE', 'COLUMN_NAME', 'NUM_BRANCH', 'OWNER']
COLUMNS_ENTRY_ORDER = ['COD_ENTITY_TARGET','COD_ENTITY_SOURCE', 'COLUMN_NAME', 'NUM_BRANCH', 'ORDER_TYPE', 'OWNER']
COLUMNS_ENTRY_FILTERS = ['COD_ENTITY_TARGET', 'VALUE', 'NUM_BRANCH', 'OWNER']
COLUMNS_ENTRY_HAVING = ['COD_ENTITY_TARGET', 'VALUE', 'NUM_BRANCH', 'OWNER']
COLUMNS_ENTRY_PATH = ['COD_PATH', 'DATABASE_NAME', 'SCHEMA_NAME', 'OWNER']
COLUMNS_ENTRY_DATASET_RELATIONSHIPS = ['COD_ENTITY_MASTER', 'COLUMN_NAME_MASTER', 'COD_ENTITY_DETAIL', 'COLUMN_NAME_DETAIL', 'RELATIONSHIP_TYPE', 'OWNER']
COLUMNS_ENTRY_ENTITY = ['COD_ENTITY', 'TABLE_NAME', 'ENTITY_TYPE', 'COD_PATH', 'STRATEGY', 'OWNER']
COLUMNS_ENTRY_DATASET_MAPPINGS = ['COD_ENTITY_SOURCE', 'VALUE_SOURCE', 'COD_ENTITY_TARGET', 'COLUMN_NAME_TARGET', 'COLUMN_TYPE_TARGET','ORDINAL_POSITION','COLUMN_LENGTH','COLUMN_PRECISION', 'NUM_BRANCH', 'KEY_TYPE', 'SW_DISTINCT', 'OWNER']
COLUMN_ENTRY_OWNER = 'OWNER'
COLUMNS_ENTRY_DV_MAPPINGS = ['COD_ENTITY_SOURCE', 'COLUMN_NAME_SOURCE','COD_ENTITY_TARGET', 'COLUMN_NAME_TARGET', 'COLUMN_TYPE_TARGET', 'ORDINAL_POSITION', 'COLUMN_LENGTH', 'COLUMN_PRECISION', 'NUM_BRANCH', 'NUM_CONNECTION', 'KEY_TYPE', 'SATELLITE_NAME', 'ORIGIN_IS_INCREMENTAL', 'ORIGIN_IS_TOTAL', 'ORIGIN_IS_CDC', 'OWNER']
COLUMNS_ENTRY_DV_ENTITY = ['COD_ENTITY', 'ENTITY_NAME', 'ENTITY_TYPE','COD_PATH', 'NAME_STATUS_TRACKING_SATELLITE', 'NAME_RECORD_TRACKING_SATELLITE', 'NAME_EFFECTIVITY_SATELLITE', 'OWNER']
COLUMNS_ENTRY_DV_PROPERTIES = ['COD_ENTITY', 'NUM_CONNECTION', 'HASH_NAME', 'OWNER']
COLUMNS_ENTRY_FILES = ['COD_ENTITY', 'FILE_PATH', 'FILE_NAME', 'DELIMITER_CHARACTER', 'OWNER']
## Git tables
COLUMNS_GIT_ENTRY_AGGREGATORS = ['COD_ENTITY_TARGET','COD_ENTITY_SOURCE', 'COLUMN_NAME', 'NUM_BRANCH', 'OWNER']
COLUMNS_GIT_ENTRY_ORDER = ['COD_ENTITY_TARGET','COD_ENTITY_SOURCE', 'COLUMN_NAME', 'NUM_BRANCH', 'ORDER_TYPE', 'OWNER']
COLUMNS_GIT_ENTRY_FILTERS = ['COD_ENTITY_TARGET', 'VALUE', 'NUM_BRANCH', 'OWNER']
COLUMNS_GIT_ENTRY_HAVING = ['COD_ENTITY_TARGET', 'VALUE', 'NUM_BRANCH', 'OWNER']
COLUMNS_GIT_ENTRY_PATH = ['COD_PATH', 'DATABASE_NAME', 'SCHEMA_NAME', 'OWNER']
COLUMNS_GIT_ENTRY_DATASET_RELATIONSHIPS = ['COD_ENTITY_MASTER', 'COLUMN_NAME_MASTER', 'COD_ENTITY_DETAIL', 'COLUMN_NAME_DETAIL', 'RELATIONSHIP_TYPE', 'OWNER']
COLUMNS_GIT_ENTRY_ENTITY = ['COD_ENTITY', 'TABLE_NAME', 'ENTITY_TYPE', 'COD_PATH', 'STRATEGY', 'OWNER']
COLUMNS_GIT_ENTRY_DATASET_MAPPINGS = ['COD_ENTITY_SOURCE', 'VALUE_SOURCE', 'COD_ENTITY_TARGET', 'COLUMN_NAME_TARGET', 'COLUMN_TYPE_TARGET','ORDINAL_POSITION','COLUMN_LENGTH','COLUMN_PRECISION', 'NUM_BRANCH', 'KEY_TYPE', 'SW_DISTINCT', 'OWNER']
COLUMNS_GIT_ENTRY_DV_ENTITY = ['COD_ENTITY', 'ENTITY_NAME', 'ENTITY_TYPE', 'COD_PATH', 'NAME_STATUS_TRACKING_SATELLITE', 'NAME_RECORD_TRACKING_SATELLITE', 'NAME_EFFECTIVITY_SATELLITE', 'OWNER']
COLUMNS_GIT_ENTRY_DV_PROPERTIES = ['COD_ENTITY', 'NUM_CONNECTION', 'HASH_NAME', 'OWNER']
COLUMNS_GIT_ENTRY_DV_MAPPINGS = ['COD_ENTITY_SOURCE', 'COLUMN_NAME_SOURCE', 'COD_ENTITY_TARGET', 'COLUMN_NAME_TARGET', 'COLUMN_TYPE_TARGET', 'ORDINAL_POSITION', 'COLUMN_LENGTH', 'COLUMN_PRECISION', 'NUM_BRANCH', 'NUM_CONNECTION', 'KEY_TYPE', 'SATELLITE_NAME', 'ORIGIN_IS_INCREMENTAL', 'ORIGIN_IS_TOTAL', 'ORIGIN_IS_CDC', 'OWNER']
COLUMNS_GIT_ENTRY_FILES = ['COD_ENTITY', 'FILE_PATH',  'FILE_NAME', 'DELIMITER_CHARACTER', 'OWNER']
COLUMN_GIT_ENTRY_OWNER = 'OWNER'
## Reference Metadata
COLUMNS_OM_REF_ORDER_TYPE = ['ID_ORDER_TYPE', 'ORDER_TYPE_NAME', 'ORDER_TYPE_VALUE', 'ORDER_TYPE_DESCRIPTION']
COLUMNS_OM_REF_MODULES = ['ID_MODULE','MODULE_NAME','MODULE_FULL_NAME', 'MODULE_DESCRIPTION']
COLUMNS_OM_REF_JOIN_TYPE = ['ID_JOIN_TYPE', 'JOIN_NAME', 'JOIN_VALUE', 'JOIN_DESCRIPTION']
COLUMNS_OM_REF_ENTITY_TYPE = ['ID_ENTITY_TYPE', 'ENTITY_TYPE_NAME', 'ENTITY_TYPE_DESCRIPTION', 'ENTITY_TYPE_FULL_NAME', 'ID_MODULE']
COLUMNS_OM_REF_KEY_TYPE = ['ID_KEY_TYPE', 'KEY_TYPE_NAME', 'KEY_TYPE_DESCRIPTION']
COLUMNS_OM_REF_QUERY_TYPE = ['ID_QUERY_TYPE', 'QUERY_TYPE_NAME', 'QUERY_TYPE_DESCRIPTION']
## Metadata
COLUMNS_OM_DATASET = ['ID_DATASET', 'DATASET_NAME', 'ID_ENTITY_TYPE', 'ID_PATH' , 'META_OWNER', 'START_DATE', 'END_DATE']
COLUMNS_OM_DATASET_EXECUTION = ['ID_DATASET', 'ID_QUERY_TYPE','META_OWNER', 'START_DATE', 'END_DATE']
COLUMNS_OM_DATASET_T_ORDER = ['ID_T_ORDER', 'ID_DATASET', 'ID_BRANCH', 'ID_DATASET_SPEC', 'ID_ORDER_TYPE','META_OWNER', 'START_DATE', 'END_DATE']
COLUMNS_OM_DATASET_T_AGG = ['ID_T_AGG', 'ID_DATASET', 'ID_BRANCH', 'ID_DATASET_SPEC','META_OWNER', 'START_DATE', 'END_DATE']
COLUMNS_OM_DATASET_T_DISTINCT = ['ID_T_DISTINCT', 'ID_DATASET', 'ID_BRANCH', 'SW_DISTINCT','META_OWNER', 'START_DATE', 'END_DATE']
COLUMNS_OM_DATASET_SPECIFICATION = ['ID_DATASET_SPEC', 'ID_DATASET', 'ID_KEY_TYPE', 'COLUMN_NAME', 'COLUMN_TYPE', 'ORDINAL_POSITION', 'IS_NULLABLE', 'COLUMN_LENGTH', 'COLUMN_PRECISION', 'COLUMN_SCALE','META_OWNER', 'START_DATE', 'END_DATE']
COLUMNS_OM_DATASET_RELATIONSHIPS = ['ID_RELATIONSHIP', 'ID_DATASET_SPEC_MASTER', 'ID_DATASET_SPEC_DETAIL', 'ID_JOIN_TYPE','META_OWNER', 'START_DATE', 'END_DATE']
COLUMNS_OM_DATASET_T_MAPPING = ['ID_T_MAPPING', 'ID_BRANCH', 'ID_DATASET_SPEC', 'VALUE_MAPPING','META_OWNER', 'START_DATE', 'END_DATE']
COLUMNS_OM_DATASET_PATH = ['ID_PATH', 'DATABASE_NAME', 'SCHEMA_NAME','META_OWNER', 'START_DATE', 'END_DATE']
COLUMNS_OM_DATASET_T_FILTER = ['ID_T_FILTER', 'ID_DATASET', 'ID_BRANCH', 'VALUE_FILTER','META_OWNER', 'START_DATE', 'END_DATE']
COLUMNS_OM_DATASET_T_HAVING = ['ID_T_HAVING', 'ID_DATASET', 'ID_BRANCH', 'VALUE_HAVING','META_OWNER', 'START_DATE', 'END_DATE']
COLUMNS_OM_DATASET_HARDCODED = ['ID_DATASET_HARDCODED', 'ID_DATASET', 'ID_BRANCH', 'CONTENT','META_OWNER', 'START_DATE', 'END_DATE']
COLUMNS_OM_PROPERTIES = ['PROPERTY', 'VALUE', 'START_DATE']
COLUMNS_OM_DATASET_DV = ['ID_DATASET', 'ID_ENTITY_TYPE', 'META_OWNER', 'START_DATE', 'END_DATE']
COLUMNS_OM_DATASET_FILE = ['ID_DATASET', 'FILE_PATH', 'FILE_NAME', 'DELIMITER_CHARACTER', 'META_OWNER', 'START_DATE', 'END_DATE']
COLUMN_OM_OWNER = 'META_OWNER'
## Information Mart
COLUMNS_OM_DATASET_INFORMATION = ['META_OWNER','ENTITY_TYPE_FULL_NAME', 'ENTITY_TYPE_DESCRIPTION', 'MODULE_NAME', 'MODULE_FULL_NAME', 'MODULE_DESCRIPTION', 'DATASET_NAME', 'DATABASE_NAME', 'SCHEMA_NAME', 'QUERY_TYPE_NAME', 'QUERY_TYPE_DESCRIPTION', 'START_DATE']
COLUMNS_OM_RELATIONSHIPS = ['META_OWNER', 'MASTER_DATABASE_NAME', 'MASTER_SCHEMA_NAME', 'MASTER_DATASET_NAME', 'MASTER_COLUMN_NAME', 'DETAIL_DATABASE_NAME', 'DETAIL_SCHEMA_NAME', 'DETAIL_DATASET_NAME', 'DETAIL_COLUMN_NAME', 'START_DATE']
COLUMNS_OM_DATASET_SPECIFICATION_INFORMATION = ['META_OWNER', 'DATASET_NAME', 'COLUMN_NAME', 'COLUMN_TYPE', 'ORDINAL_POSITION', 'IS_NULLABLE', 'COLUMN_LENGTH', "COLUMN_PRECISION", "COLUMN_SCALE"]

# ENGINES
## Entities
ENTITY_SRC = "SRC"
ENTITY_TB = "TB"
ENTITY_WITH = "WITH"
ENTITY_VIEW = "VIEW"
ENTITY_HUB = "HUB"
ENTITY_LINK = "LINK"
ENTITY_SAT = "SAT"
ENTITY_STS = "STS"
ENTITY_RTS = "RTS"
ENTITY_SATE = "SATE"

## Key Types
KEY_TYPE_PRIMARY_KEY = "PK"
KEY_TYPE_NULL = "NULL"
KEY_TYPE_BUSINESS_KEY = "BK"
KEY_TYPE_SEQUENCE = "SEQ"
KEY_TYPE_FOREIGN_KEY = "FK"
KEY_TYPE_HASH_KEY = "HK"
KEY_TYPE_ATTRIBUTE = "AT"
KEY_TYPE_STATUS = "ST"
KEY_TYPE_DRIVENKEY = "DK"
KEY_TYPE_RECORD_SOURCE = "RS"
KEY_TYPE_TENANT = "TN"
KEY_TYPE_HASHDIFF = "HD"
KEY_TYPE_DEPENDENT_CHILD_KEY = "DCK"
KEY_TYPE_APPLIED_DATE = "AD"

## Status Nodes
NODE_STATUS_WAITING = "WAITING"
NODE_STATUS_RUNNING = "RUNNING"
NODE_STATUS_FINISHED_OK = "OK"
NODE_STATUS_FINISHED_NOK = "NOK"
NODE_STATUS_SKIP = "SKIP"

# OUTPUT
##Output Types
OUTPUT_TYPE_SQL = "sql"
OUTPUT_TYPE_DBT = "dbt"


from metamorf.engines.engine import Engine
from metamorf.tools.filecontroller import FileControllerFactory
from metamorf.tools.connection import ConnectionFactory
from metamorf.constants import *
from metamorf.tools.metadata import Metadata
from metamorf.tools.query import Query
import os

class EngineUpload(Engine):

    def _initialize_engine(self):
        self.engine_name = "Engine Upload"
        self.engine_command = "upload"

        self.entry_files_to_load = [FILE_ENTRY_AGGREGATORS, FILE_ENTRY_DATASET_MAPPINGS, FILE_ENTRY_DATASET_RELATIONSHIPS,
                                    FILE_ENTRY_ENTITY, FILE_ENTRY_FILTERS, FILE_ENTRY_ORDER, FILE_ENTRY_PATH,
                                    FILE_ENTRY_HAVING, FILE_ENTRY_DV_ENTITY, FILE_ENTRY_DV_PROPERTIES,
                                    FILE_ENTRY_DV_MAPPINGS, FILE_ENTRY_FILES]
        if 'select' in self.arguments:
            if self.arguments['select'] == "all" or self.arguments['select'] == "*":
                self.entry_files_to_load = [FILE_ENTRY_AGGREGATORS, FILE_ENTRY_DATASET_MAPPINGS, FILE_ENTRY_DATASET_RELATIONSHIPS,
                                            FILE_ENTRY_ENTITY, FILE_ENTRY_FILTERS, FILE_ENTRY_ORDER, FILE_ENTRY_PATH,
                                            FILE_ENTRY_HAVING, FILE_ENTRY_DV_ENTITY, FILE_ENTRY_DV_PROPERTIES,
                                            FILE_ENTRY_DV_MAPPINGS, FILE_ENTRY_FILES]
            if self.arguments['select'].lower() == FILE_ENTRY_DATASET_MAPPINGS.lower():
                self.entry_files_to_load = [FILE_ENTRY_DATASET_MAPPINGS]
            if self.arguments['select'].lower() == FILE_ENTRY_DATASET_RELATIONSHIPS.lower():
                self.entry_files_to_load = [FILE_ENTRY_DATASET_RELATIONSHIPS]
            if self.arguments['select'].lower() == FILE_ENTRY_ENTITY.lower():
                self.entry_files_to_load = [FILE_ENTRY_ENTITY]
            if self.arguments['select'].lower() == FILE_ENTRY_FILTERS.lower():
                self.entry_files_to_load = [FILE_ENTRY_FILTERS]
            if self.arguments['select'].lower() == FILE_ENTRY_ORDER.lower():
                self.entry_files_to_load = [FILE_ENTRY_ORDER]
            if self.arguments['select'].lower() == FILE_ENTRY_PATH.lower():
                self.entry_files_to_load = [FILE_ENTRY_PATH]
            if self.arguments['select'].lower() == FILE_ENTRY_HAVING.lower():
                self.entry_files_to_load = [FILE_ENTRY_HAVING]
            if self.arguments['select'].lower() == FILE_ENTRY_DV_ENTITY.lower():
                self.tables_to_load = [FILE_ENTRY_DV_ENTITY]
            if self.arguments['select'].lower() == FILE_ENTRY_DV_PROPERTIES.lower():
                self.tables_to_load = [FILE_ENTRY_DV_PROPERTIES]
            if self.arguments['select'].lower() == FILE_ENTRY_DV_MAPPINGS.lower():
                self.tables_to_load = [FILE_ENTRY_DV_MAPPINGS]
            if self.arguments['select'].lower() == FILE_ENTRY_FILES.lower():
                self.tables_to_load = [FILE_ENTRY_FILES]


    def run(self):
        # Starts the execution loading the Configuration File. If there is an error it finishes the execution.
        super().start_execution()
        connection_type = self.configuration_file['metadata']['connection_type']
        self.connection = ConnectionFactory().get_connection(connection_type)
        self.connection.setup_connection(self.configuration_file['metadata'], self.log)

        result_delete = self.delete_all_entry_from_owner()
        if not result_delete:
            self.connection.rollback()
            super().finish_execution(result_delete)
        else:
            result = self.load_files_to_metadata_database()
            if not result: self.connection.rollback()
            else: self.connection.commit()
            super().finish_execution(result)

    def delete_all_entry_from_owner(self):
        result = True
        self.log.log(self.engine_name, "Starting deleting Metadata Entry on database" , LOG_LEVEL_INFO)
        connection_type = self.configuration_file['metadata']['connection_type']
        for file in self.entry_files_to_load:
            self.log.log(self.engine_name, "Deleting: " + file, LOG_LEVEL_INFO)
            query = Query()
            query.set_database(connection_type)
            query.set_type(QUERY_TYPE_DELETE)
            query.set_target_table(file)
            query.set_where_filters([COLUMN_ENTRY_OWNER + "='" + self.owner + "'"])
            res = self.connection.execute(str(query))
            result = result and res
        self.log.log(self.engine_name, "Finished deleting Metadata Entry on database", LOG_LEVEL_INFO)
        return result

    def add_owner_at_end_each_row(self, file: list[list]):
        result = []
        for f in file:
            s = f
            s.append(self.configuration_file['owner'])
            result.append(s)
        return result

    def load_files_to_metadata_database(self):
        self.log.log(self.engine_name, "Starting uploading Metadata Entry on database", LOG_LEVEL_INFO)
        result = True
        metadata = Metadata(self.log)
        connection_type = self.configuration_file['metadata']['connection_type']

        # ENTRY_AGGREGATORS
        if FILE_ENTRY_AGGREGATORS in self.entry_files_to_load:
            file_reader = FileControllerFactory().get_file_reader(FILE_TYPE_CSV)
            file_reader.set_file_location(os.path.join(ACTUAL_PATH, ENTRY_FILES_PATH),
                                          FILE_ENTRY_AGGREGATORS + "." + FILE_TYPE_CSV)
            file = file_reader.read_file()

            all_rows = self.add_owner_at_end_each_row(file)
            res = metadata.add_entry_aggregators(all_rows)
            result = result and res
            query_values = Query()
            query_values.set_database(connection_type)
            query_values.set_has_header(True)
            query_values.set_type(QUERY_TYPE_VALUES)
            query_values.set_target_table(TABLE_ENTRY_AGGREGATORS)
            query_values.set_insert_columns(COLUMNS_ENTRY_AGGREGATORS)
            query_values.set_values(metadata.entry_aggregators)
            self.log.log(self.engine_name, "Loading: " + TABLE_ENTRY_AGGREGATORS, LOG_LEVEL_INFO)
            res = self.connection.execute(str(query_values))
            result = result and res

        # ENTRY_DATASET_MAPPINGS
        if FILE_ENTRY_DATASET_MAPPINGS in self.entry_files_to_load:
            file_reader = FileControllerFactory().get_file_reader(FILE_TYPE_CSV)
            file_reader.set_file_location(os.path.join(ACTUAL_PATH, ENTRY_FILES_PATH),
                                          FILE_ENTRY_DATASET_MAPPINGS + "." + FILE_TYPE_CSV)
            file = file_reader.read_file()
            all_rows = self.add_owner_at_end_each_row(file)
            res = metadata.add_entry_dataset_mappings(all_rows)
            result = result and res
            query_values = Query()
            query_values.set_database(connection_type)
            query_values.set_has_header(True)
            query_values.set_type(QUERY_TYPE_VALUES)
            query_values.set_target_table(TABLE_ENTRY_DATASET_MAPPINGS)
            query_values.set_insert_columns(COLUMNS_ENTRY_DATASET_MAPPINGS)
            query_values.set_values(metadata.entry_dataset_mappings)
            self.log.log(self.engine_name, "Loading: " + TABLE_ENTRY_DATASET_MAPPINGS, LOG_LEVEL_INFO)
            res = self.connection.execute(str(query_values))
            result = result and res

        # ENTRY_DATASET_RELATIONSHIPS
        if FILE_ENTRY_DATASET_RELATIONSHIPS in self.entry_files_to_load:
            file_reader = FileControllerFactory().get_file_reader(FILE_TYPE_CSV)
            file_reader.set_file_location(os.path.join(ACTUAL_PATH, ENTRY_FILES_PATH),
                                          FILE_ENTRY_DATASET_RELATIONSHIPS + "." + FILE_TYPE_CSV)
            file = file_reader.read_file()
            all_rows = self.add_owner_at_end_each_row(file)
            res = metadata.add_entry_dataset_relationship(all_rows)
            result = result and res
            query_values = Query()
            query_values.set_database(connection_type)
            query_values.set_has_header(True)
            query_values.set_type(QUERY_TYPE_VALUES)
            query_values.set_target_table(TABLE_ENTRY_DATASET_RELATIONSHIPS)
            query_values.set_insert_columns(COLUMNS_ENTRY_DATASET_RELATIONSHIPS)
            query_values.set_values(metadata.entry_dataset_relationship)
            self.log.log(self.engine_name, "Loading: " + TABLE_ENTRY_DATASET_RELATIONSHIPS, LOG_LEVEL_INFO)
            res = self.connection.execute(str(query_values))
            result = result and res

        # ENTRY_ENTITY
        if FILE_ENTRY_ENTITY in self.entry_files_to_load:
            file_reader = FileControllerFactory().get_file_reader(FILE_TYPE_CSV)
            file_reader.set_file_location(os.path.join(ACTUAL_PATH, ENTRY_FILES_PATH), FILE_ENTRY_ENTITY + "." + FILE_TYPE_CSV)
            file = file_reader.read_file()
            all_rows = self.add_owner_at_end_each_row(file)
            res = metadata.add_entry_entity(all_rows)
            result = result and res
            query_values = Query()
            query_values.set_database(connection_type)
            query_values.set_has_header(True)
            query_values.set_type(QUERY_TYPE_VALUES)
            query_values.set_target_table(TABLE_ENTRY_ENTITY)
            query_values.set_insert_columns(COLUMNS_ENTRY_ENTITY)
            query_values.set_values(metadata.entry_entity)
            self.log.log(self.engine_name, "Loading: " + TABLE_ENTRY_ENTITY, LOG_LEVEL_INFO)
            res = self.connection.execute(str(query_values))
            result = result and res

        # ENTRY_FILTERS
        if FILE_ENTRY_FILTERS in self.entry_files_to_load:
            file_reader = FileControllerFactory().get_file_reader(FILE_TYPE_CSV)
            file_reader.set_file_location(os.path.join(ACTUAL_PATH, ENTRY_FILES_PATH), FILE_ENTRY_FILTERS + "." + FILE_TYPE_CSV)
            file = file_reader.read_file()
            all_rows = self.add_owner_at_end_each_row(file)
            res = metadata.add_entry_filters(all_rows)
            result = result and res
            query_values = Query()
            query_values.set_database(connection_type)
            query_values.set_has_header(True)
            query_values.set_type(QUERY_TYPE_VALUES)
            query_values.set_target_table(TABLE_ENTRY_FILTERS)
            query_values.set_insert_columns(COLUMNS_ENTRY_FILTERS)
            query_values.set_values(metadata.entry_filters)
            self.log.log(self.engine_name, "Loading: " + TABLE_ENTRY_FILTERS, LOG_LEVEL_INFO)
            res = self.connection.execute(str(query_values))
            result = result and res

        # ENTRY_HAVING
        if FILE_ENTRY_HAVING in self.entry_files_to_load:
            file_reader = FileControllerFactory().get_file_reader(FILE_TYPE_CSV)
            file_reader.set_file_location(os.path.join(ACTUAL_PATH, ENTRY_FILES_PATH),
                                          FILE_ENTRY_HAVING + "." + FILE_TYPE_CSV)
            file = file_reader.read_file()
            all_rows = self.add_owner_at_end_each_row(file)
            res = metadata.add_entry_having(all_rows)
            result = result and res
            query_values = Query()
            query_values.set_database(connection_type)
            query_values.set_has_header(True)
            query_values.set_type(QUERY_TYPE_VALUES)
            query_values.set_target_table(TABLE_ENTRY_HAVING)
            query_values.set_insert_columns(COLUMNS_ENTRY_HAVING)
            query_values.set_values(metadata.entry_having)
            self.log.log(self.engine_name, "Loading: " + TABLE_ENTRY_HAVING, LOG_LEVEL_INFO)
            res = self.connection.execute(str(query_values))
            result = result and res

        # ENTRY_ORDER
        if FILE_ENTRY_ORDER in self.entry_files_to_load:
            file_reader = FileControllerFactory().get_file_reader(FILE_TYPE_CSV)
            file_reader.set_file_location(os.path.join(ACTUAL_PATH, ENTRY_FILES_PATH), FILE_ENTRY_ORDER + "." + FILE_TYPE_CSV)
            file = file_reader.read_file()
            all_rows = self.add_owner_at_end_each_row(file)
            res = metadata.add_entry_order(all_rows)
            result = result and res
            query_values = Query()
            query_values.set_database(connection_type)
            query_values.set_has_header(True)
            query_values.set_type(QUERY_TYPE_VALUES)
            query_values.set_target_table(TABLE_ENTRY_ORDER)
            query_values.set_insert_columns(COLUMNS_ENTRY_ORDER)
            query_values.set_values(metadata.entry_order)
            self.log.log(self.engine_name, "Loading: " + TABLE_ENTRY_ORDER, LOG_LEVEL_INFO)
            res = self.connection.execute(str(query_values))
            result = result and res

        # ENTRY_PATH
        if FILE_ENTRY_PATH in self.entry_files_to_load:
            file_reader = FileControllerFactory().get_file_reader(FILE_TYPE_CSV)
            file_reader.set_file_location(os.path.join(ACTUAL_PATH, ENTRY_FILES_PATH), FILE_ENTRY_PATH + "." + FILE_TYPE_CSV)
            file = file_reader.read_file()
            all_rows = self.add_owner_at_end_each_row(file)
            res = metadata.add_entry_path(all_rows)
            result = result and res
            query_values = Query()
            query_values.set_database(connection_type)
            query_values.set_has_header(True)
            query_values.set_type(QUERY_TYPE_VALUES)
            query_values.set_target_table(TABLE_ENTRY_PATH)
            query_values.set_insert_columns(COLUMNS_ENTRY_PATH)
            query_values.set_values(metadata.entry_path)
            self.log.log(self.engine_name, "Loading: " + TABLE_ENTRY_PATH, LOG_LEVEL_INFO)
            res = self.connection.execute(str(query_values))
            result = result and res

        # ENTRY_DV_ENTITY
        if FILE_ENTRY_DV_ENTITY in self.entry_files_to_load:
            file_reader = FileControllerFactory().get_file_reader(FILE_TYPE_CSV)
            file_reader.set_file_location(os.path.join(ACTUAL_PATH, ENTRY_FILES_PATH),
                                          FILE_ENTRY_DV_ENTITY + "." + FILE_TYPE_CSV)
            file = file_reader.read_file()
            all_rows = self.add_owner_at_end_each_row(file)
            res = metadata.add_entry_dv_entity(all_rows)
            result = result and res
            query_values = Query()
            query_values.set_database(connection_type)
            query_values.set_has_header(True)
            query_values.set_type(QUERY_TYPE_VALUES)
            query_values.set_target_table(TABLE_ENTRY_DV_ENTITY)
            query_values.set_insert_columns(COLUMNS_ENTRY_DV_ENTITY)
            query_values.set_values(metadata.entry_dv_entity)
            self.log.log(self.engine_name, "Loading: " + TABLE_ENTRY_DV_ENTITY, LOG_LEVEL_INFO)
            res = self.connection.execute(str(query_values))
            result = result and res

        # ENTRY_DV_MAPPINGS
        if FILE_ENTRY_DV_MAPPINGS in self.entry_files_to_load:
            file_reader = FileControllerFactory().get_file_reader(FILE_TYPE_CSV)
            file_reader.set_file_location(os.path.join(ACTUAL_PATH, ENTRY_FILES_PATH), FILE_ENTRY_DV_MAPPINGS + "." + FILE_TYPE_CSV)
            file = file_reader.read_file()
            all_rows = self.add_owner_at_end_each_row(file)
            res = metadata.add_entry_dv_mappings(all_rows)
            result = result and res
            query_values = Query()
            query_values.set_database(connection_type)
            query_values.set_has_header(True)
            query_values.set_type(QUERY_TYPE_VALUES)
            query_values.set_target_table(TABLE_ENTRY_DV_MAPPINGS)
            query_values.set_insert_columns(COLUMNS_ENTRY_DV_MAPPINGS)
            query_values.set_values(metadata.entry_dv_mappings)
            self.log.log(self.engine_name, "Loading: " + TABLE_ENTRY_DV_MAPPINGS, LOG_LEVEL_INFO)
            res = self.connection.execute(str(query_values))
            result = result and res

        # ENTRY_DV_PROPERTIES
        if FILE_ENTRY_DV_PROPERTIES in self.entry_files_to_load:
            file_reader = FileControllerFactory().get_file_reader(FILE_TYPE_CSV)
            file_reader.set_file_location(os.path.join(ACTUAL_PATH, ENTRY_FILES_PATH),
                                          FILE_ENTRY_DV_PROPERTIES + "." + FILE_TYPE_CSV)
            file = file_reader.read_file()
            all_rows = self.add_owner_at_end_each_row(file)
            res = metadata.add_entry_dv_properties(all_rows)
            result = result and res
            query_values = Query()
            query_values.set_database(connection_type)
            query_values.set_has_header(True)
            query_values.set_type(QUERY_TYPE_VALUES)
            query_values.set_target_table(TABLE_ENTRY_DV_PROPERTIES)
            query_values.set_insert_columns(COLUMNS_ENTRY_DV_PROPERTIES)
            query_values.set_values(metadata.entry_dv_properties)
            self.log.log(self.engine_name, "Loading: " + TABLE_ENTRY_DV_PROPERTIES, LOG_LEVEL_INFO)
            res = self.connection.execute(str(query_values))
            result = result and res

        # ENTRY_FILES
        if FILE_ENTRY_FILES in self.entry_files_to_load:
            file_reader = FileControllerFactory().get_file_reader(FILE_TYPE_CSV)
            file_reader.set_file_location(os.path.join(ACTUAL_PATH, ENTRY_FILES_PATH),
                                          FILE_ENTRY_FILES + "." + FILE_TYPE_CSV)
            file = file_reader.read_file()
            all_rows = self.add_owner_at_end_each_row(file)
            res = metadata.add_entry_files(all_rows)
            result = result and res
            query_values = Query()
            query_values.set_database(connection_type)
            query_values.set_has_header(True)
            query_values.set_type(QUERY_TYPE_VALUES)
            query_values.set_target_table(TABLE_ENTRY_FILES)
            query_values.set_insert_columns(COLUMNS_ENTRY_FILES)
            query_values.set_values(metadata.entry_files)
            self.log.log(self.engine_name, "Loading: " + TABLE_ENTRY_FILES, LOG_LEVEL_INFO)
            res = self.connection.execute(str(query_values))
            result = result and res

        self.log.log(self.engine_name, "Finished uploading Metadata Entry on database", LOG_LEVEL_INFO)
        return result



from metamorf.tools.database_objects import *
from metamorf.tools.query import Query, FromRelationQuery, OrderByQuery
from metamorf.constants import *
from metamorf.tools.metadata import Metadata
from metamorf.tools.connection import Connection
from metamorf.tools.log import Log
from metamorf.tools.node import Node
import re

def Sort_Dataset_Specification_By_Ordinal_Position(sub_list):
    sub_list.sort(key=lambda x: x.ordinal_position)
    return sub_list

def get_metadata_from_database(connection: Connection, log: Log, owner: str, load_om: bool=True, load_ref: bool=True, load_entry: bool=True, load_im: bool=True):

    metadata = Metadata(log)

    query = Query()
    query.set_type(QUERY_TYPE_SELECT)
    log.log("Metadata Loader", "Starting to load Metadata into metamorf", LOG_LEVEL_INFO)

    if load_om:
        where_filter = COLUMN_OM_OWNER + "='" + owner + "'"
        query.set_where_filters(where_filter)

        # OM_DATASET
        log.log("Metadata Loader", "Start to load [OM_DATASET]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_OM_DATASET)
        query.set_select_columns(COLUMNS_OM_DATASET)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res= metadata.add_om_dataset(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [OM_DATASET]", LOG_LEVEL_DEBUG)

        # OM_DATASET_DV
        log.log("Metadata Loader", "Start to load [OM_DATASET_DV]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_OM_DATASET_DV)
        query.set_select_columns(COLUMNS_OM_DATASET_DV)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_om_dataset_dv(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [OM_DATASET_DV]", LOG_LEVEL_DEBUG)

        # OM_DATASET_EXECUTION
        log.log("Metadata Loader", "Start to load [OM_DATASET_EXECUTION]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_OM_DATASET_EXECUTION)
        query.set_select_columns(COLUMNS_OM_DATASET_EXECUTION)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_om_dataset_execution(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [OM_DATASET_EXECUTION]", LOG_LEVEL_DEBUG)

        # OM_DATASET_T_ORDER
        log.log("Metadata Loader", "Start to load [OM_DATASET_T_ORDER]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_OM_DATASET_T_ORDER)
        query.set_select_columns(COLUMNS_OM_DATASET_T_ORDER)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_om_dataset_t_order(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [OM_DATASET_T_ORDER]", LOG_LEVEL_DEBUG)

        # OM_DATASET_T_AGG
        log.log("Metadata Loader", "Start to load [OM_DATASET_T_AGG]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_OM_DATASET_T_AGG)
        query.set_select_columns(COLUMNS_OM_DATASET_T_AGG)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_om_dataset_t_agg(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [OM_DATASET_T_AGG]", LOG_LEVEL_DEBUG)

        # OM_DATASET_T_DISTINCT
        log.log("Metadata Loader", "Start to load [OM_DATASET_T_DISTINCT]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_OM_DATASET_T_DISTINCT)
        query.set_select_columns(COLUMNS_OM_DATASET_T_DISTINCT)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_om_dataset_t_distinct(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [OM_DATASET_T_DISTINCT]", LOG_LEVEL_DEBUG)

        # OM_DATASET_T_FILTER
        log.log("Metadata Loader", "Start to load [OM_DATASET_T_FILTER]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_OM_DATASET_T_FILTER)
        query.set_select_columns(COLUMNS_OM_DATASET_T_FILTER)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_om_dataset_t_filter(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [OM_DATASET_T_FILTER]", LOG_LEVEL_DEBUG)

        # OM_DATASET_T_HAVING
        log.log("Metadata Loader", "Start to load [OM_DATASET_T_HAVING]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_OM_DATASET_T_HAVING)
        query.set_select_columns(COLUMNS_OM_DATASET_T_HAVING)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_om_dataset_t_having(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [OM_DATASET_T_HAVING]", LOG_LEVEL_DEBUG)

        # OM_DATASET_T_MAPPING
        log.log("Metadata Loader", "Start to load [OM_DATASET_T_MAPPING]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_OM_DATASET_T_MAPPING)
        query.set_select_columns(COLUMNS_OM_DATASET_T_MAPPING)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_om_dataset_t_mapping(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [OM_DATASET_T_MAPPING]", LOG_LEVEL_DEBUG)

        # OM_DATASET_PATH
        log.log("Metadata Loader", "Start to load [OM_DATASET_PATH]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_OM_DATASET_PATH)
        query.set_select_columns(COLUMNS_OM_DATASET_PATH)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_om_dataset_path(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [OM_DATASET_PATH]", LOG_LEVEL_DEBUG)

        # OM_DATASET_SPECIFICATION
        log.log("Metadata Loader", "Start to load [OM_DATASET_SPECIFICATION]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_OM_DATASET_SPECIFICATION)
        query.set_select_columns(COLUMNS_OM_DATASET_SPECIFICATION)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_om_dataset_specification(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [OM_DATASET_SPECIFICATION]", LOG_LEVEL_DEBUG)

        # OM_DATASET_RELATIONSHIPS
        log.log("Metadata Loader", "Start to load [OM_DATASET_RELATIONSHIPS]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_OM_DATASET_RELATIONSHIPS)
        query.set_select_columns(COLUMNS_OM_DATASET_RELATIONSHIPS)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_om_dataset_relationships(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [OM_DATASET_RELATIONSHIPS]", LOG_LEVEL_DEBUG)

        # OM_DATASET_FILE
        log.log("Metadata Loader", "Start to load [OM_DATASET_FILE]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_OM_DATASET_FILE)
        query.set_select_columns(COLUMNS_OM_DATASET_FILE)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_om_dataset_file(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [OM_DATASET_FILE]", LOG_LEVEL_DEBUG)

        # OM_DATASET_HARDCODED
        log.log("Metadata Loader", "Start to load [OM_DATASET_HARDCODED]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_OM_DATASET_HARDCODED)
        query.set_select_columns(COLUMNS_OM_DATASET_HARDCODED)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_om_dataset_hardcoded(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [OM_DATASET_HARDCODED]", LOG_LEVEL_DEBUG)

        # OM_PROPERTIES
        log.log("Metadata Loader", "Start to load [OM_PROPERTIES]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_OM_PROPERTIES)
        query.set_select_columns(COLUMNS_OM_PROPERTIES)
        query.set_where_filters(None)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_om_properties(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [OM_PROPERTIES]", LOG_LEVEL_DEBUG)


    if load_ref:
        query.set_where_filters(None)

        # OM_REF_QUERY_TYPE
        log.log("Metadata Loader", "Start to load [OM_REF_QUERY_TYPE]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_OM_REF_QUERY_TYPE)
        query.set_select_columns(COLUMNS_OM_REF_QUERY_TYPE)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_om_ref_query_type(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [OM_REF_QUERY_TYPE]", LOG_LEVEL_DEBUG)

        # OM_REF_ORDER_TYPE
        log.log("Metadata Loader", "Start to load [OM_REF_ORDER_TYPE]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_OM_REF_ORDER_TYPE)
        query.set_select_columns(COLUMNS_OM_REF_ORDER_TYPE)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_om_ref_order_type(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [OM_REF_ORDER_TYPE]", LOG_LEVEL_DEBUG)

        # OM_REF_JOIN_TYPE
        log.log("Metadata Loader", "Start to load [OM_REF_JOIN_TYPE]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_OM_REF_JOIN_TYPE)
        query.set_select_columns(COLUMNS_OM_REF_JOIN_TYPE)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_om_ref_join_type(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [OM_REF_JOIN_TYPE]", LOG_LEVEL_DEBUG)

        # OM_REF_MODULES
        log.log("Metadata Loader", "Start to load [OM_REF_MODULES]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_OM_REF_MODULES)
        query.set_select_columns(COLUMNS_OM_REF_MODULES)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_om_ref_modules(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [OM_REF_MODULES]", LOG_LEVEL_DEBUG)

        # OM_REF_ENTITY_TYPE
        log.log("Metadata Loader", "Start to load [OM_REF_ENTITY_TYPE]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_OM_REF_ENTITY_TYPE)
        query.set_select_columns(COLUMNS_OM_REF_ENTITY_TYPE)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_om_ref_entity_type(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [OM_REF_ENTITY_TYPE]", LOG_LEVEL_DEBUG)

        # OM_REF_KEY_TYPE
        log.log("Metadata Loader", "Start to load [OM_REF_KEY_TYPE]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_OM_REF_KEY_TYPE)
        query.set_select_columns(COLUMNS_OM_REF_KEY_TYPE)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_om_ref_key_type(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [OM_REF_KEY_TYPE]", LOG_LEVEL_DEBUG)

    if load_entry:
        where_filter = COLUMN_ENTRY_OWNER + "='" + owner + "'"
        query.set_where_filters(where_filter)

        # ENTRY_ORDER
        log.log("Metadata Loader", "Start to load [ENTRY_ORDER]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_ENTRY_ORDER)
        query.set_select_columns(COLUMNS_ENTRY_ORDER)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_entry_order(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [ENTRY_ORDER]", LOG_LEVEL_DEBUG)

        # ENTRY_AGGREGATORS
        log.log("Metadata Loader", "Start to load [ENTRY_AGGREGATORS]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_ENTRY_AGGREGATORS)
        query.set_select_columns(COLUMNS_ENTRY_AGGREGATORS)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_entry_aggregators(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [ENTRY_AGGREGATORS]", LOG_LEVEL_DEBUG)

        # ENTRY_FILTERS
        log.log("Metadata Loader", "Start to load [ENTRY_FILTERS]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_ENTRY_FILTERS)
        query.set_select_columns(COLUMNS_ENTRY_FILTERS)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_entry_filters(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [ENTRY_FILTERS]", LOG_LEVEL_DEBUG)

        # ENTRY_HAVING
        log.log("Metadata Loader", "Start to load [ENTRY_FILTERS]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_ENTRY_HAVING)
        query.set_select_columns(COLUMNS_ENTRY_HAVING)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_entry_having(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [ENTRY_FILTERS]", LOG_LEVEL_DEBUG)

        # ENTRY_DATASET_RELATIONSHIPS
        log.log("Metadata Loader", "Start to load [ENTRY_DATASET_RELATIONSHIPS]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_ENTRY_DATASET_RELATIONSHIPS)
        query.set_select_columns(COLUMNS_ENTRY_DATASET_RELATIONSHIPS)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_entry_dataset_relationship(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [ENTRY_DATASET_RELATIONSHIPS]", LOG_LEVEL_DEBUG)

        # ENTRY_ENTITY
        log.log("Metadata Loader", "Start to load [ENTRY_ENTITY]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_ENTRY_ENTITY)
        query.set_select_columns(COLUMNS_ENTRY_ENTITY)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_entry_entity(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [ENTRY_ENTITY]", LOG_LEVEL_DEBUG)

        # ENTRY_PATH
        log.log("Metadata Loader", "Start to load [ENTRY_PATH]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_ENTRY_PATH)
        query.set_select_columns(COLUMNS_ENTRY_PATH)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_entry_path(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [ENTRY_PATH]", LOG_LEVEL_DEBUG)

        # ENTRY_DATASET_MAPPINGS
        log.log("Metadata Loader", "Start to load [ENTRY_DATASET_MAPPINGS]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_ENTRY_DATASET_MAPPINGS)
        query.set_select_columns(COLUMNS_ENTRY_DATASET_MAPPINGS)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_entry_dataset_mappings(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [ENTRY_DATASET_MAPPINGS]", LOG_LEVEL_DEBUG)

        # ENTRY_DV_MAPPINGS
        log.log("Metadata Loader", "Start to load [ENTRY_DV_MAPPINGS]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_ENTRY_DV_MAPPINGS)
        query.set_select_columns(COLUMNS_ENTRY_DV_MAPPINGS)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_entry_dv_mappings(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [ENTRY_DV_MAPPINGS]", LOG_LEVEL_DEBUG)

        # ENTRY_DV_ENTITY
        log.log("Metadata Loader", "Start to load [TABLE_ENTRY_DV_ENTITY]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_ENTRY_DV_ENTITY)
        query.set_select_columns(COLUMNS_ENTRY_DV_ENTITY)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_entry_dv_entity(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [TABLE_ENTRY_DV_ENTITY]", LOG_LEVEL_DEBUG)

        # ENTRY_DV_PROPERTIES
        log.log("Metadata Loader", "Start to load [TABLE_ENTRY_DV_PROPERTIES]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_ENTRY_DV_PROPERTIES)
        query.set_select_columns(COLUMNS_ENTRY_DV_PROPERTIES)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_entry_dv_properties(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [TABLE_ENTRY_DV_PROPERTIES]", LOG_LEVEL_DEBUG)

        # ENTRY_FILES
        log.log("Metadata Loader", "Start to load [TABLE_ENTRY_FILES]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_ENTRY_FILES)
        query.set_select_columns(COLUMNS_ENTRY_FILES)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_entry_files(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [TABLE_ENTRY_FILES]", LOG_LEVEL_DEBUG)

    if load_im:
        where_filter = COLUMN_OM_OWNER + "='" + owner + "'"
        query.set_where_filters(where_filter)

        # OM_DATASET_INFORMATION
        log.log("Metadata Loader", "Start to load [OM_DATASET_INFORMATION]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_IM_OM_DATASET_INFORMATION)
        query.set_select_columns(COLUMNS_OM_DATASET_INFORMATION)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_om_dataset_information(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [OM_DATASET_INFORMATION]", LOG_LEVEL_DEBUG)

        # OM_RELATIONSHIPS
        log.log("Metadata Loader", "Start to load [OM_RELATIONSHIPS]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_IM_OM_RELATIONSHIPS)
        query.set_select_columns(COLUMNS_OM_RELATIONSHIPS)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_om_relationships(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [OM_RELATIONSHIPS]", LOG_LEVEL_DEBUG)

        # OM_DATASET_SPECIFICATION_INFORMATION
        log.log("Metadata Loader", "Start to load [OM_DATASET_SPECIFICATION_INFORMATION]", LOG_LEVEL_DEBUG)
        query.set_from_tables(TABLE_IM_OM_DATASET_SPECIFICATION_INFORMATION)
        query.set_select_columns(COLUMNS_OM_DATASET_SPECIFICATION_INFORMATION)
        res = connection.execute(str(query))
        if res is False: return None
        data = connection.get_query_result()
        res = metadata.add_om_dataset_specification_information(data)
        if res is False: return None
        log.log("Metadata Loader", "Finished to load [OM_DATASET_SPECIFICATION_INFORMATION]", LOG_LEVEL_DEBUG)



    log.log("Metadata Loader", "Finished to load Metadata into metamorf", LOG_LEVEL_INFO)
    return metadata

def get_all_sources_and_withs_from_dataset(metadata, dataset):
    # Get Sources
    all_sources = metadata.get_all_source_tables_from_target_table_name(dataset.dataset_name)
    for src in all_sources:
        dataset_src = metadata.get_dataset_from_dataset_name(src)
        if dataset_src.id_entity_type == metadata.get_entity_type_from_entity_type_name(ENTITY_WITH).id_entity_type:
            all_new_sources = get_all_sources_and_withs_from_dataset(metadata, dataset_src)
            for x in all_new_sources:
                x_src = metadata.get_dataset_from_dataset_name(x)
                if x_src.id_entity_type == metadata.get_entity_type_from_entity_type_name(ENTITY_WITH).id_entity_type:
                    all_sources.append(x)


    return all_sources

def get_list_nodes_from_metadata(metadata: Metadata, log: Log, dataset_name: str):

    log.log("Metadata Nodes", "Start to load all the Nodes", LOG_LEVEL_INFO)
    all_nodes = []
    for dataset in [x for x in metadata.om_dataset if x.end_date is None]:
        if dataset.id_entity_type == metadata.get_entity_type_from_entity_type_name(ENTITY_WITH).id_entity_type or dataset.id_entity_type == metadata.get_entity_type_from_entity_type_name(ENTITY_SRC).id_entity_type: continue
        # Get all sources from dataset: separated on tables/views - withs
        all_sources_names = get_all_sources_and_withs_from_dataset(metadata, dataset)
        from_tables = []
        with_tables = []
        for src in all_sources_names:
            data = metadata.get_dataset_from_dataset_name(src)
            if data.id_entity_type != metadata.get_entity_type_from_entity_type_name(ENTITY_WITH).id_entity_type:
                from_tables.append(src)
            else:
                with_tables.append(src)
                with_to_get_source = metadata.get_dataset_from_dataset_name(src)
                from_tables.extend(get_sources_from_dataset_with(metadata, with_to_get_source))

        if dataset.dataset_name in from_tables: from_tables = [i for i in from_tables if i != dataset.dataset_name]
        from_tables = list(dict.fromkeys(from_tables))


        main_query = get_query_object_from_dataset(dataset, False, metadata)

        for with_table in list(set(with_tables)):
            with_dataset = metadata.get_dataset_from_dataset_name(with_table)
            with_query = get_query_object_from_dataset(with_dataset, True, metadata)
            main_query.add_subquery(with_query)

        all_nodes.append(Node(name=dataset.dataset_name, query=main_query, predecessors=from_tables))

    log.log("Metadata Nodes", "Finished to load the Nodes", LOG_LEVEL_INFO)
    return all_nodes

def get_sources_from_dataset_with(metadata,dataset):
    all_sources_names = get_all_sources_and_withs_from_dataset(metadata, dataset)
    new_sources = []
    for src in all_sources_names:
        data = metadata.get_dataset_from_dataset_name(src)
        if data.id_entity_type != metadata.get_entity_type_from_entity_type_name(ENTITY_WITH).id_entity_type:
            new_sources.append(src)
        else:
            new_sources.extend(get_sources_from_dataset_with(metadata, data))
    return new_sources

def get_query_object_from_dataset(dataset: OmDataset, is_with: bool, metadata: Metadata):

    num_branches = metadata.get_num_branches_from_dataset_name(dataset.dataset_name)
    # INSERT Columns
    insert_columns = []
    for dataset_spec in Sort_Dataset_Specification_By_Ordinal_Position([x for x in metadata.om_dataset_specification if x.end_date is None and x.id_dataset == dataset.id_dataset]):
        insert_columns.append(dataset_spec.column_name)

    # PK Columns
    pk_columns = []
    for dataset_spec in [x for x in metadata.om_dataset_specification if x.end_date is None and x.id_dataset == dataset.id_dataset and x.id_key_type == metadata.get_key_type_from_key_type_name(KEY_TYPE_PRIMARY_KEY)]:
        pk_columns.append(metadata.get_tables_dot_column_from_id_dataset_spec(dataset_spec.id_dataset_spec))

    # Create table
    columns_specs = []
    all_dataset_specs_from_dataset = [x for x in metadata.om_dataset_specification if x.end_date is None and x.id_dataset == dataset.id_dataset]
    all_dataset_specs_from_dataset.sort(key=lambda x: x.ordinal_position)
    for dataset_spec in all_dataset_specs_from_dataset:
        length_details = ''
        if dataset_spec.column_length != 0 and dataset_spec.column_length is not None:
            length_details = "(" + str(dataset_spec.column_length) + ")"
        if dataset_spec.column_precision != 0 and dataset_spec.column_precision is not None:
            if dataset_spec.column_scale != 0 and dataset_spec.column_scale is not None:
                length_details = "(" + str(dataset_spec.column_scale) + ","+str(dataset_spec.column_scale)+")"
            else:
                length_details = "(" + str(dataset_spec.column_scale)+ ")"
        columns_specs.append(dataset_spec.column_name + " " + dataset_spec.column_type+length_details)
    query = Query()

    for num_branch in range(0,num_branches):

        from_tables = metadata.get_all_source_tables_from_target_table_name_and_branch(dataset.dataset_name, num_branch+1)
        # Sources and Relationships
        tables_and_relations = []
        for rel in metadata.om_dataset_relationships:
            if rel.end_date is not None: continue
            dataset_spec_master = metadata.get_dataset_spec_from_id_dataset_spec(rel.id_dataset_spec_master)
            dataset_master = metadata.get_dataset_from_id_dataset(dataset_spec_master.id_dataset)
            dataset_spec_detail = metadata.get_dataset_spec_from_id_dataset_spec(rel.id_dataset_spec_detail)
            dataset_detail = metadata.get_dataset_from_id_dataset(dataset_spec_detail.id_dataset)

            if dataset_master.id_entity_type != metadata.get_entity_type_from_entity_type_name(ENTITY_WITH).id_entity_type:
                new_dataset_master = metadata.get_dataset_fqdn_from_dataset_name(dataset_master.dataset_name)
            else:
                new_dataset_master = dataset_master.dataset_name

            if dataset_detail.id_entity_type != metadata.get_entity_type_from_entity_type_name(ENTITY_WITH).id_entity_type:
                new_dataset_detail = metadata.get_dataset_fqdn_from_dataset_name(dataset_detail.dataset_name)
            else:
                new_dataset_detail = dataset_detail.dataset_name

            if new_dataset_master in from_tables and new_dataset_detail in from_tables:
                tables_and_relations.append(
                    FromRelationQuery(new_dataset_master, dataset_spec_master.column_name,
                                      new_dataset_detail, dataset_spec_detail.column_name, rel.id_join_type,
                                      "="))

        # If the source is added on the relationships, is it removed from the FROM Tables
        for relations in tables_and_relations:
            if relations.master_table in from_tables:
                from_tables.remove(relations.master_table)
            if relations.detail_table in from_tables:
                from_tables.remove(relations.detail_table)

        # SELECT Columns
        select_columns = []
        new_dataset = metadata.get_dataset_from_dataset_name(dataset.dataset_name)
        for dataset_spec in Sort_Dataset_Specification_By_Ordinal_Position([x for x in metadata.om_dataset_specification if x.end_date is None and new_dataset.id_dataset == x.id_dataset]):
            dataset_t_mapping = metadata.get_dataset_t_mapping_from_id_branch_and_id_dataset_spec(num_branch+1, dataset_spec.id_dataset_spec)
            if dataset_t_mapping is None: continue # If the mapping on that branch has been closed
            new_select_column = dataset_t_mapping.value_mapping
            all_columns = re.findall(r'\[(\d+)\]', new_select_column)
            for col in all_columns:
                new_select_column = new_select_column.replace("[" + col + "]",
                                                              metadata.get_tables_dot_column_from_id_dataset_spec(col))
            new_select_column += ' AS ' + metadata.get_dataset_spec_from_id_dataset_spec(dataset_t_mapping.id_dataset_spec).column_name
            select_columns.append(new_select_column)


        # DISTINCT option
        is_distinct = False
        for distinct in [x for x in metadata.om_dataset_t_distinct if x.end_date is None]:
            if distinct.id_dataset == dataset.id_dataset and distinct.id_branch == (num_branch+1):
                if distinct.sw_distinct == 1: is_distinct = True

        # ORDER Columns
        all_id_dataset_spec = metadata.get_list_id_dataset_spec_from_dataset_name(dataset.dataset_name)
        order_columns = []
        for order in [x for x in metadata.om_dataset_t_order if x.end_date is None]:
            if order.id_dataset_spec in all_id_dataset_spec and order.id_branch == (num_branch+1):
                order_type = metadata.get_order_type_from_id_order_type(order.id_order_type)
                order_columns.append(
                    OrderByQuery(metadata.get_tables_dot_column_from_id_dataset_spec(order.id_dataset_spec),
                                 order_type.order_type_value))

        # AGG Columns
        agg_columns = []
        for agg in [x for x in metadata.om_dataset_t_agg if x.end_date is None]:
            if agg.id_dataset == dataset.id_dataset and agg.id_branch == (num_branch+1):
                agg_columns.append(metadata.get_tables_dot_column_from_id_dataset_spec(agg.id_dataset_spec))

        # WHERE Filters
        where_filters = []
        for filter in [x for x in metadata.om_dataset_t_filter if x.end_date is None]:
            if filter.id_dataset == dataset.id_dataset and filter.id_branch == (num_branch+1):
                filter_value = filter.value_filter
                all_columns = re.findall(r'\[(\d+)\]', filter.value_filter)
                for col in all_columns:
                    filter_value = filter_value.replace("[" + col + "]",
                                                        metadata.get_tables_dot_column_from_id_dataset_spec(col))
                where_filters.append(filter_value)

        # HAVING Filters
        having_filters = []
        for having in [x for x in metadata.om_dataset_t_having if x.end_date is None]:
            if having.id_dataset == dataset.id_dataset and having.id_branch == (num_branch+1):
                having_value = having.value_having
                all_columns = re.findall(r'\[(\d+)\]', having.value_having)
                for col in all_columns:
                    having_value = having_value.replace("[" + col + "]",
                                                        metadata.get_tables_dot_column_from_id_dataset_spec(col))
                having_filters.append(having_value)

        if not is_with:
            target_table = metadata.get_dataset_fqdn_from_dataset_name(dataset.dataset_name)
        else:
            target_table = dataset.dataset_name



        if num_branch == 0:
            query.set_name_query(target_table)
            query.set_target_table(target_table)
            query.set_columns_and_specs(columns_specs)
            query.set_is_with(is_with)
            if is_with: query.set_type(QUERY_TYPE_SELECT)
            query.set_select_columns(select_columns)
            query.set_from_tables(from_tables)
            query.set_is_distinct(is_distinct)
            query.set_from_tables_and_relations(tables_and_relations)
            query.set_insert_columns(insert_columns)
            query.set_order_by_columns(order_columns)
            query.set_having_filters(having_filters)
            query.set_group_by_columns(agg_columns)
            query.set_primary_key(pk_columns)
            query.set_where_filters(where_filters)
        else:

            query_union = Query()
            query_union.set_name_query(target_table)
            query_union.set_target_table(target_table)
            query_union.set_is_with(False)
            query_union.set_select_columns(select_columns)
            query_union.set_from_tables(from_tables)
            query_union.set_is_distinct(is_distinct)
            query_union.set_from_tables_and_relations(tables_and_relations)
            query_union.set_order_by_columns(order_columns)
            query_union.set_having_filters(having_filters)
            query_union.set_group_by_columns(agg_columns)
            query_union.set_primary_key(pk_columns)
            query_union.set_type(QUERY_TYPE_SELECT)
            query_union.set_where_filters(where_filters)
            if len(select_columns) != 0:
                query.add_unionquery(query_union)


    return query

def get_node_with_with_query_execution_settings(connection, metadata, node):
    ''' Returns the node object with query execution settings configured '''
    does_table_exists = connection.does_table_exists(node.name)
    dataset = metadata.get_dataset_from_dataset_name(node.name)
    id_query_type = metadata.get_dataset_execution_from_id_dataset(dataset.id_dataset).id_query_type
    node.query.set_database(connection.get_connection_type())

    if id_query_type == QUERY_TYPE_INSERT:
        node.query.set_need_create_table(not does_table_exists)
        node.set_node_type(QUERY_TYPE_INSERT)
    if id_query_type == QUERY_TYPE_VIEW:
        node.set_node_type(QUERY_TYPE_VIEW)
    if does_table_exists and id_query_type == QUERY_TYPE_TRUNCATE_AND_INSERT:
        node.set_node_type(QUERY_TYPE_INSERT)
        node.query.set_is_truncate(True)
    if not does_table_exists and id_query_type == QUERY_TYPE_TRUNCATE_AND_INSERT:
        node.set_node_type(QUERY_TYPE_INSERT)
        node.query.set_need_create_table(True)
        node.query.set_is_truncate(True)

    return node


from metamorf.engines.engine import Engine
from metamorf.tools.filecontroller import FileControllerFactory
from metamorf.tools.connection import ConnectionFactory
from metamorf.constants import *
from metamorf.tools.metadata import Metadata
from metamorf.tools.query import Query
import glob
import os

class EngineFiles(Engine):

    def _initialize_engine(self):
        self.engine_name = "Engine Files"
        self.engine_command = "files"


    def run(self):
        # Starts the execution loading the Configuration File. If there is an error it finishes the execution.
        super().start_execution()
        self.connection_type = self.configuration_file['data']['connection_type']
        self.connection_data = ConnectionFactory().get_connection(self.configuration_file['data']['connection_type'])
        self.connection_metadata = ConnectionFactory().get_connection(self.configuration_file['metadata']['connection_type'])
        self.connection_data.setup_connection(self.configuration_file['data'], self.log)
        self.connection_metadata.setup_connection(self.configuration_file['metadata'], self.log)

        self.metadata_actual = self.load_metadata(load_om=False, load_entry=True, load_ref=False, load_im=False, owner=self.owner)
        self.metadata_to_load = Metadata(self.log)

        result = self.load_files_to_data_database()
        super().finish_execution(result)


    def load_files_to_data_database(self):
        self.log.log(self.engine_name, "Starting to upload files", LOG_LEVEL_INFO)
        result = True
        for file in self.metadata_actual.entry_files:

            self.log.log(self.engine_name, "Starting to upload ['"+str(os.path.join(file.file_path , file.file_name))+"'] file", LOG_LEVEL_INFO)
            entry_entity = self.metadata_actual.get_entry_entity_from_cod_entity(file.cod_entity)

            file_reader = FileControllerFactory().get_file_reader(FILE_TYPE_CSV)
            file_reader.set_delimiter(file.delimiter_character)

            file_list = glob.glob(os.path.join(file.file_path,file.file_name))

            all_rows = []
            header = []
            is_header = True
            first_time = False
            for f in file_list:
                file_reader.set_file_location(None, f)
                file_content = file_reader.read_file()
                if file_content is None:
                    self.log.log(self.engine_name, "The file ['" + str(os.path.join(file.file_path , file.file_name)) + "'] has not been found", LOG_LEVEL_WARNING)
                    return result
                for row in file_content:
                    if first_time:
                        first_time = False
                        continue
                    if is_header:
                        header = row
                        is_header = False
                    new_row = ''
                    for pos in row:
                        new_row += "'" + pos.replace("'","''") + "',"
                    all_rows.append(new_row[:-1])
                first_time = True

            columns_and_specs = []
            for r in header:
                new_column = r + " " + self.connection_data.get_string_for_metadata()[0]
                columns_and_specs.append(new_column)

            table_exists = self.connection_data.does_table_exists(entry_entity.table_name)

            query_values = Query()
            query_values.set_need_drop_table(table_exists)
            query_values.set_need_create_table(True)
            query_values.set_database(self.connection_type)
            query_values.set_has_header(True)
            query_values.set_type(QUERY_TYPE_VALUES)
            query_values.set_target_table(entry_entity.table_name)
            query_values.set_insert_columns(header)
            query_values.set_columns_and_specs(columns_and_specs)
            query_values.set_values(all_rows)

            self.connection_data.execute(str(query_values))
            self.connection_data.commit()

            self.log.log(self.engine_name, "The file ['" + str(os.path.join(file.file_path, file.file_name)) + "'] has been uploaded", LOG_LEVEL_INFO)

        self.log.log(self.engine_name, "Finished to upload files", LOG_LEVEL_INFO)

        return result
